from math import pi, sqrt, cos, sin
from cmath import phase
import cmath

meson_quark = { 'B0': 'bd', 'Bs': 'bs', 'K0': 'sd', 'D0': 'cu' }

def bag_msbar2rgi(alpha_s, meson):
    """Conversion factor between renormalization group invariant (RGI) defintion
    $\hat B$ of the bag parameter and the $\overline{\mathrm{MS}}$ definition
    $B(mu)$:
    $$\hat B = b_B^{(n_f)} B(mu)$$

    See e.g. eq. (84) in arXiv:1011.4408.
    """
    J={}
    if meson in ['B0', 'Bs']: # nf=5
        J = 5165/3174.
        g = 6/23
    elif meson == 'K0': # nf=3
        J = 307/162.
        g = 2/9.
    return alpha_s**(-g) * (1 + alpha_s/(4*pi) * J)

def DeltaM(M12, G12):
    r"""Meson mixing mass difference $\Delta M$ as a function of $M_{12}$ and
    $\Gamma_{12}$."""
    aM12 = abs(M12)
    aG12 = abs(G12)
    phi12 = phase(-M12/G12)
    return sqrt(4*aM12**2 + aG12**2 + sqrt(16*aM12**4 + 16*aM12**2*aG12**2
                    + aG12**4 + 8*aM12**2*aG12**2*cos(2*phi12)))/sqrt(2)

def DeltaGamma(M12, G12):
    r"""Meson mixing decay width difference $\Delta\Gamma$ as a function of
    $M_{12}$ and $\Gamma_{12}$."""
    aM12 = abs(M12)
    aG12 = abs(G12)
    phi12 = phase(-M12/G12)
    return sqrt(2)*cmath.sqrt(-4*aM12**2 - aG12**2 +
                sqrt(16*aM12**4 + 8*aM12**2*aG12**2 +
                    aG12**4 + 16*aM12**2*aG12**2*cos(phi12)**2)).real

def q_over_p(M12, G12):
    r"""Ratio $q/p$ as a function of $M_{12}$ and $\Gamma_{12}$."""
    DM = DeltaM(M12, G12)
    DG = DeltaGamma(M12, G12)
    return -cmath.sqrt((2*M12.conjugate()-1j*G12.conjugate())/(2*M12-1j*G12))

def a_fs(M12, G12):
    r"""Flavour-specific CP asymmetry in meson mixing as a function of
    $M_{12}$ and $\Gamma_{12}$."""
    aM12 = abs(M12)
    aG12 = abs(G12)
    phi12 = phase(-M12/G12)
    return aG12 / aM12 * sin(phi12)



"""Special mathematical functions"""


import scipy
import mpmath

def zeta(x):
    """Riemann Zeta function"""
    return scipy.special.zeta(x, 1)

def li2(x):
    r"""Complex Dilogarithm"""
    # for real x<=1: use the special scipy function which is 20x faster
    if x.imag==0. and x.real <= 1:
        return scipy.special.spence(1-x.real)
    return mpmath.fp.polylog(2,x)

def ei(x):
    """Exponential integral function"""
    return scipy.special.expi(x)



import pkgutil
import csv
import numpy as np
from flavio.classes import Parameter
from flavio.statistics.probability import MultivariateNormalDistribution

def csv_to_dict(filename):
    f = pkgutil.get_data('flavio.physics', filename)
    datareader = csv.reader(f.decode('utf-8').splitlines(), dialect='excel-tab')
    res = {}
    for line in datareader:
        if len(line) == 2: # for the central values
            # do not read the results for the c parameters - they are not needed.
            if line[0].split('_')[1][0]=='c':
                continue
            res[line[0]] = float(line[1])
        elif len(line) == 3: # for the covariance
            # do not read the results for the c parameters - they are not needed.
            if line[0].split('_')[1][0]=='c' or line[1].split('_')[1][0]=='c':
                continue
            res[(line[0],line[1])] = float(line[2])
    return res


def load_parameters(file_res, file_cov, process, constraints):
    implementation_name = process + ' SSE'
    res_dict = csv_to_dict(file_res)
    cov_dict = csv_to_dict(file_cov)
    keys_sorted = sorted(res_dict.keys())
    res = [res_dict[k] for k in keys_sorted]
    # M -> M + M^T - diag(M) since the dictionary contains only the entries above the diagonal
    cov = ( np.array([[ cov_dict.get((k,m),0) for m in keys_sorted] for k in keys_sorted])
          + np.array([[ cov_dict.get((m,k),0) for m in keys_sorted] for k in keys_sorted])
          - np.diag([ cov_dict[(k,k)] for k in keys_sorted]) )
    parameter_names = [implementation_name + ' ' + coeff_name for coeff_name in keys_sorted]
    for parameter_name in parameter_names:
        try: # check if parameter object already exists
            p = Parameter.get_instance(parameter_name)
        except: # otherwise, create a new one
            p = Parameter(parameter_name)
        else: # if parameter exists, remove existing constraints
            constraints.remove_constraints(parameter_name)
    constraints.add_constraint(parameter_names,
            MultivariateNormalDistribution(central_value=res, covariance=cov ))


def lattice_load(constraints):
    load_parameters('data/arXiv-1501-00367v2/av_sl_results.d',
                    'data/arXiv-1501-00367v2/av_sl_covariance.d',
                    'B->K*', constraints)
    load_parameters('data/arXiv-1501-00367v2/av_ls_results.d',
                    'data/arXiv-1501-00367v2/av_ls_covariance.d',
                    'Bs->K*', constraints)
    load_parameters('data/arXiv-1501-00367v2/av_ss_results.d',
                    'data/arXiv-1501-00367v2/av_ss_covariance.d',
                    'Bs->phi', constraints)
    load_parameters('data/arXiv-1501-00367v2/t_sl_results.d',
                    'data/arXiv-1501-00367v2/t_sl_covariance.d',
                    'B->K*', constraints)
    load_parameters('data/arXiv-1501-00367v2/t_ls_results.d',
                    'data/arXiv-1501-00367v2/t_ls_covariance.d',
                    'Bs->K*', constraints)
    load_parameters('data/arXiv-1501-00367v2/t_ss_results.d',
                    'data/arXiv-1501-00367v2/t_ss_covariance.d',
                    'Bs->phi', constraints)



"""A fit is a collection of observables and parameters that can be used to
perform statistical analyses within a particular statistical framework.

Fits are instances of descendants of the `Fit` class (which is not meant
to be used directly)."""

import flavio
import numpy as np
import copy

class Fit(flavio.NamedInstanceClass):
    """Base class for fits. Not meant to be used directly."""

    def __init__(self,
                 name,
                 par_obj,
                 fit_parameters,
                 nuisance_parameters,
                 observables,
                 fit_wc_names=[],
                 fit_wc_function=None,
                 fit_wc_priors=None,
                 input_scale=160.,
                 exclude_measurements=None,
                ):
        # some checks to make sure the input is sane
        for p in fit_parameters + nuisance_parameters:
            # check that fit and nuisance parameters exist
            assert p in par_obj._parameters.keys(), "Parameter " + p + " not found in Constraints"
        for obs in observables:
            # check that observables exist
            try:
                if isinstance(obs, tuple):
                    flavio.classes.Observable.get_instance(obs[0])
                else:
                    flavio.classes.Observable.get_instance(obs)
            except:
                raise ValueError("Observable " + str(obs) + " not found!")
        # check that no parameter appears as fit *and* nuisance parameter
        intersect = set(fit_parameters).intersection(nuisance_parameters)
        assert intersect == set(), "Parameters appearing as fit_parameters and nuisance_parameters: " + str(intersect)
        # check that the Wilson coefficient function works
        if fit_wc_names: # if list of WC names not empty
            try:
                fit_wc_function(**{fit_wc_name: 1e-6 for fit_wc_name in fit_wc_names})
            except:
                raise ValueError("Error in calling the Wilson coefficient function")
        # now that everything seems fine, we can call the init of the parent class
        super().__init__(name)
        self.par_obj = par_obj
        self.parameters_central = self.par_obj.get_central_all()
        self.fit_parameters = fit_parameters
        self.nuisance_parameters = nuisance_parameters
        self.exclude_measurements = exclude_measurements
        self.fit_wc_names = fit_wc_names
        self.fit_wc_function = fit_wc_function
        self.fit_wc_priors = fit_wc_priors
        self.observables = observables
        self.input_scale = input_scale

    @property
    def get_central_fit_parameters(self):
        """Return a numpy array with the central values of all fit parameters."""
        return np.asarray([self.parameters_central[p] for p in self.fit_parameters])

    @property
    def get_random_fit_parameters(self):
        """Return a numpy array with random values for all fit parameters."""
        all_random = self.par_obj.get_random_all()
        return np.asarray([all_random[p] for p in self.fit_parameters])

    @property
    def get_random_wilson_coeffs(self):
        """Return a numpy array with random values for all Wilson coefficients."""
        if self.fit_wc_priors is None:
            return None
        all_random = self.fit_wc_priors.get_random_all()
        return np.asarray([all_random[p] for p in self.fit_wc_names])

    @property
    def get_central_nuisance_parameters(self):
        """Return a numpy array with the central values of all nuisance parameters."""
        return np.asarray([self.parameters_central[p] for p in self.nuisance_parameters])

    @property
    def get_random_nuisance_parameters(self):
        """Return a numpy array with random values for all nuisance parameters."""
        all_random = self.par_obj.get_random_all()
        return np.asarray([all_random[p] for p in self.nuisance_parameters])

    @property
    def get_measurements(self):
        """Return a list of all the measurements currently defined that
        constrain any of the fit observables."""
        all_measurements = []
        for m_name, m_obj in flavio.classes.Measurement.instances.items():
            if set(m_obj.all_parameters).isdisjoint(self.observables):
                # if set of all observables constrained by measurement is disjoint
                # with fit observables, do nothing
                continue
            else:
                # else, add measurement name to output list
                all_measurements.append(m_name)
        if self.exclude_measurements is None:
            return all_measurements
        else:
            return list(set(all_measurements) - set(self.exclude_measurements))



class BayesianFit(Fit):
    """Bayesian fit class. Instances of this class can then be fed to samplers.

    Parameters
    ----------

    - `name`: a descriptive string name
    - `par_obj`: an instance of `ParameterConstraints`, e.g. `flavio.default_parameters`
    - `fit_parameters`: a list of string names of parameters of interest. The existing
      constraints on the parameter will be taken as prior.
    - `nuisance_parameters`: a list of string names of nuisance parameters. The existing
      constraints on the parameter will be taken as prior.
    - `observables`: a list of observable names to be included in the fit
    - `exclude_measurements`: a list of measurement names not to be included in
    the fit
    - `fit_wc_names`: optional; a list of string names of arguments of the Wilson
      coefficient function below
    - `fit_wc_function`: optional; a function that has exactly the arguements listed
      in `fit_wc_names` and returns a dictionary that can be fed to the `set_initial`
      method of the Wilson coefficient class. Example: fit the real and imaginary
      parts of $C_{10}$ in $b\to s\mu^+\mu^-$.
    ```
    def fit_wc_function(Re_C10, Im_C10):
        return {'C10_bsmmumu': Re_C10 + 1j*Im_C10}
    ```
    - `input_scale`: input scale for the Wilson coeffficients. Defaults to 160.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dimension = len(self.fit_parameters) + len(self.nuisance_parameters) + len(self.fit_wc_names)

    def array_to_dict(self, x):
        """Convert a 1D numpy array of floats to a dictionary of fit parameters,
        nuisance parameters, and Wilson coefficients."""
        d = {}
        n_fit_p = len(self.fit_parameters)
        n_nui_p = len(self.nuisance_parameters)
        n_wc = len(self.fit_wc_names)
        d['fit_parameters'] = { p: x[i] for i, p in enumerate(self.fit_parameters) }
        d['nuisance_parameters'] = { p: x[i + n_fit_p] for i, p in enumerate(self.nuisance_parameters) }
        d['fit_wc'] = { p: x[i + n_fit_p + n_nui_p] for i, p in enumerate(self.fit_wc_names) }
        return d

    def dict_to_array(self, d):
        """Convert a dictionary of fit parameters,
        nuisance parameters, and Wilson coefficients to a 1D numpy array of
        floats."""
        n_fit_p = len(self.fit_parameters)
        n_nui_p = len(self.nuisance_parameters)
        n_wc = len(self.fit_wc_names)
        arr = np.zeros(n_fit_p + n_nui_p + n_wc)
        arr[:n_fit_p] = [d['fit_parameters'][p] for p in self.fit_parameters]
        arr[n_fit_p:n_fit_p+n_nui_p] = [d['nuisance_parameters'][p] for p in self.nuisance_parameters]
        arr[n_fit_p+n_nui_p:]   = [d['fit_wc'][c] for c in self.fit_wc_names]
        return arr

    @property
    def get_random(self):
        """Get an array with random values for all the fit and nuisance
        parameters"""
        arr = np.zeros(self.dimension)
        n_fit_p = len(self.fit_parameters)
        n_nui_p = len(self.nuisance_parameters)
        arr[:n_fit_p] = self.get_random_fit_parameters
        arr[n_fit_p:n_fit_p+n_nui_p] = self.get_random_nuisance_parameters
        arr[n_fit_p+n_nui_p:] = self.get_random_wilson_coeffs
        return arr

    def get_par_dict(self, x):
        """Get a dictionary of fit and nuisance parameters from an input array"""
        d = self.array_to_dict(x)
        par_dict = self.parameters_central.copy()
        par_dict.update(d['fit_parameters'])
        par_dict.update(d['nuisance_parameters'])
        return par_dict

    def get_wc_obj(self, x):
        wc_obj = flavio.WilsonCoefficients()
        # if there are no WCs to be fitted, return the SM WCs
        if not self.fit_wc_names:
            return wc_obj
        d = self.array_to_dict(x)
        wc_obj.set_initial(self.fit_wc_function(**d['fit_wc']), self.input_scale)
        return wc_obj

    def log_prior_parameters(self, x):
        """Return the prior probability for all fit and nuisance parameters
        given an input array"""
        par_dict = self.get_par_dict(x)
        exclude_parameters = list(set(par_dict.keys())-set(self.fit_parameters)-set(self.nuisance_parameters))
        prob_dict = self.par_obj.get_logprobability_all(par_dict, exclude_parameters=exclude_parameters)
        return sum([p for obj, p in prob_dict.items()])

    def log_prior_wilson_coeffs(self, x):
        """Return the prior probability for all Wilson coefficients
        given an input array"""
        if self.fit_wc_priors is None:
            return 0
        wc_dict = self.array_to_dict(x)['fit_wc']
        prob_dict = self.fit_wc_priors.get_logprobability_all(wc_dict)
        return sum([p for obj, p in prob_dict.items()])

    def get_predictions(self, x):
        """Get a dictionary with predictions for all observables given an input
        array"""
        par_dict = self.get_par_dict(x)
        wc_obj = self.get_wc_obj(x)
        all_predictions = {}
        for observable in self.observables:
            if isinstance(observable, tuple):
                obs_name = observable[0]
                _inst = flavio.classes.Observable.get_instance(obs_name)
                all_predictions[observable] = _inst.prediction_par(par_dict, wc_obj, *observable[1:])
            else:
                _inst = flavio.classes.Observable.get_instance(observable)
                all_predictions[observable] = _inst.prediction_par(par_dict, wc_obj)
        return all_predictions

    def log_likelihood(self, x):
        """Return the logarithm of the likelihood function (not including the
        prior)"""
        predictions = self.get_predictions(x)
        ll = 0.
        for measurement in self.get_measurements:
            m_obj = flavio.Measurement.get_instance(measurement)
            m_obs = m_obj.all_parameters
            exclude_observables = set(m_obs) - set(self.observables)
            prob_dict = m_obj.get_logprobability_all(predictions, exclude_parameters=exclude_observables)
            ll += sum(prob_dict.values())
        return ll

    def log_target(self, x):
        """Return the logarithm of the likelihood times prior probability"""
        return self.log_likelihood(x) + self.log_prior_parameters(x) + self.log_prior_wilson_coeffs(x)


class FastFit(BayesianFit):
    """A subclass of `BayesianFit` that is meant to produce fast likelihood
    contour plots.

    Calling the method `make_measurement`, a pseudo-measurement is generated
    that combines the actual experimental measurements with the theoretical
    uncertainties stemming from the nuisance parameters. This is done by
    generating random samples of the nuisance parameters and evaluating all
    observables within the Standard Model many times (100 by default).
    Then, the covariance of all predictions is extracted. Similarly, a covariance
    matrix for all experimental measurements is determined. Both covariance
    matrices are added and the resulting multivariate Gaussian treated as a
    single measurement.

    This approach has the advantage that two-dimensional plots of the likelihood
    can be produced without the need for sampling or profiling the other
    dimensions. However, several strong assumptions go into this method, most
    importantly,

    - all uncertainties - experimental and theoretical - are treated as Gaussian
    - the theoretical uncertainties in the presence of new physics are assumed
      to be similar to the ones in the SM
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.measurements = None


    # a method to get the mean and covariance of all measurements of all
    # observables of interest
    def _get_central_covariance_experiment(self, N=100):
        random_dict = {}
        for obs in self.observables:
            # intialize empty lists
            random_dict[obs] = []
        for measurement in self.get_measurements:
            if measurement.split(' ')[0] == 'Pseudo-measurement':
                continue
            m_obj = flavio.Measurement.get_instance(measurement)
            for i in range(N):
                m_random = m_obj.get_random_all()
                m_obs = m_obj.all_parameters
                our_obs = set(m_obs).intersection(self.observables)
                for obs in our_obs:
                    random_dict[obs].append(m_random[obs])
        random_arr = np.zeros((len(self.observables), N))
        for i, obs in enumerate(self.observables):
            n = len(random_dict[obs])
            random_arr[i] = random_dict[obs][::n//N]
        return np.mean(random_arr, axis=1), np.cov(random_arr)

    # a method to get the covariance of the SM prediction of all observables
    # of interest
    def _get_covariance_sm(self, N=100):
        par_central = self.par_obj.get_central_all()
        def random_nuisance_dict():
            arr = self.get_random_nuisance_parameters
            nuis_dict = {par: arr[i] for i, par in enumerate(self.nuisance_parameters)}
            par = par_central.copy()
            par.update(nuis_dict)
            return par
        par_random = [random_nuisance_dict() for i in range(N)]

        pred_arr = np.zeros((len(self.observables), N))
        wc_sm = flavio.WilsonCoefficients()
        for i, observable in enumerate(self.observables):
            if isinstance(observable, tuple):
                obs_name = observable[0]
                _inst = flavio.classes.Observable.get_instance(obs_name)
                pred_arr[i] = np.array([_inst.prediction_par(par, wc_sm, *observable[1:])
                                        for par in par_random])
            else:
                _inst = flavio.classes.Observable.get_instance(observable)
                pred_arr[i] = np.array([_inst.prediction_par(par, wc_sm)
                                        for par in par_random])
        return np.cov(pred_arr)

    def make_measurement(self, N=100):
        """Initialize the fit by producing a pseudo-measurement containing both
        experimental uncertainties as well as theory uncertainties stemming
        from nuisance parameters."""
        central_exp, cov_exp = self._get_central_covariance_experiment(N)
        cov_sm = self._get_covariance_sm(N)
        covariance = cov_exp + cov_sm
        if not len(central_exp) == 1:
            if not np.all(np.linalg.eigvals(covariance) > 0):
                # if the covariance matrix is not positive definite, try a dirty trick:
                # multiply all the correlations by 0.9.
                n_dim = len(covariance)
                err = np.sqrt(np.diag(covariance))
                correlation = covariance/err/err[:,np.newaxis] # correlation matrix
                correlation = (correlation - np.eye(n_dim))*0.9 + np.eye(n_dim)
                covariance = np.outer(err, err) * correlation
        # add the Pseudo-measurement
        m = flavio.classes.Measurement('Pseudo-measurement for FastFit instance: ' + self.name)
        if len(central_exp) == 1:
            m.add_constraint(self.observables,
                    flavio.statistics.probability.NormalDistribution(central_exp, np.sqrt(covariance)))
        else:
            m.add_constraint(self.observables,
                    flavio.statistics.probability.MultivariateNormalDistribution(central_exp, covariance))

    def array_to_dict(self, x):
        """Convert a 1D numpy array of floats to a dictionary of fit parameters,
        nuisance parameters, and Wilson coefficients."""
        d = {}
        n_fit_p = len(self.fit_parameters)
        n_wc = len(self.fit_wc_names)
        d['fit_parameters'] = { p: x[i] for i, p in enumerate(self.fit_parameters) }
        d['fit_wc'] = { p: x[i + n_fit_p] for i, p in enumerate(self.fit_wc_names) }
        return d

    def dict_to_array(self, d):
        """Convert a dictionary of fit parameters and Wilson coefficients to a
        1D numpy array of floats."""
        n_fit_p = len(self.fit_parameters)
        n_wc = len(self.fit_wc_names)
        arr = np.zeros(n_fit_p + n_nui_p + n_wc)
        arr[:n_fit_p] = [d['fit_parameters'][p] for p in self.fit_parameters]
        arr[n_fit_p:]   = [d['fit_wc'][c] for c in self.fit_wc_names]
        return arr

    def get_par_dict(self, x):
        d = self.array_to_dict(x)
        par_dict = self.parameters_central.copy()
        par_dict.update(d['fit_parameters'])
        return par_dict

    def log_likelihood(self, x):
        """Return the logarithm of the likelihood. Note that there is no prior
        probability for nuisance parameters, which have been integrated out.
        Priors for fit parameters are ignored."""
        predictions = self.get_predictions(x)
        m_obj = flavio.Measurement.get_instance('Pseudo-measurement for FastFit instance: ' + self.name)
        m_obs = m_obj.all_parameters
        prob_dict = m_obj.get_logprobability_all(predictions)
        ll = sum(prob_dict.values())
        return ll



"""Functions for imroved Isgur-Wise relations to obtain tensor
form factors from vector form factors in the heavy quark limit."""

import flavio
from math import pi,log


def improved_isgur_wise(process, q2, ff, par, scale):
    pd = flavio.physics.bdecays.formfactors.b_v.cln.process_dict[process]
    mB = par['m_'+pd['B']]
    mV = par['m_'+pd['V']]
    mb = flavio.physics.running.running.get_mb(par, scale)
    alpha_s = flavio.physics.running.running.get_alpha(par, scale)['alpha_s']
    D0v = 0
    C0v = 0
    # eq. (3.8) of arXiv:1006.5013
    kappa = 1-2*alpha_s/(3*pi)*log(scale/mb)
    # power corrections
    a_T1  = par[process + ' IW a_T1']
    a_T2  = par[process + ' IW a_T2']
    a_T23 = par[process + ' IW a_T23']
    # eq. (3.6) of arXiv:1006.5013
    ff['T1'] = kappa*ff['V']  * ( 1 + a_T1 )
    ff['T2'] = kappa*ff['A1'] * ( 1 + a_T2 )
    # derived in analogy to arXiv:1006.5013 using hep-ph/0404250
    ff['T23'] = kappa*ff['A12'] * 2*mB**2/q2 * ( 1 + a_T23 )
    return ff



import unittest
import numpy as np
from .bvgamma import *
from flavio.physics.bdecays.formfactors.b_v import bsz_parameters
from flavio.physics.eft import WilsonCoefficients
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.parameters import default_parameters
import flavio


wc = WilsonCoefficients()
par = default_parameters#

class TestBVgamma(unittest.TestCase):
    def test_bksgamma(self):
        # just check if this works
        flavio.Observable.get_instance("ACP(B0->K*gamma)").prediction_central(par, wc)
        flavio.Observable.get_instance("S_K*gamma").prediction_central(par, wc)
        # numerical comparison to  David's old Mathematica code
        self.assertAlmostEqual(
            flavio.Observable.get_instance("BR(B0->K*gamma)").prediction_central(par, wc)*1e5/3.91526,
             1, places=1)
        self.assertAlmostEqual(
            flavio.Observable.get_instance("BR(B+->K*gamma)").prediction_central(par, wc)*1e5/4.11625,
             1, places=1)

    def test_bksgamma(self):
        # just check if this works
        flavio.Observable.get_instance("ACP(Bs->phigamma)").prediction_central(par, wc)
        flavio.Observable.get_instance("S_phigamma").prediction_central(par, wc)
        flavio.Observable.get_instance("BR(Bs->phigamma)").prediction_central(par, wc)
        flavio.Observable.get_instance("ADeltaGamma(Bs->phigamma)").prediction_central(par, wc)



import unittest
import numpy as np
import flavio
from math import pi, sqrt, exp, log

class TestProbability(unittest.TestCase):
    def test_multiv_normal(self):
        # test that the rescaling of the MultivariateNormalDistribution
        # does not affect the log PDF!
        c = np.array([1e-3, 2])
        cov = np.array([[(0.2e-3)**2, 0.2e-3*0.5*0.3],[0.2e-3*0.5*0.3, 0.5**2]])
        pdf = flavio.statistics.probability.MultivariateNormalDistribution(c, cov)
        x=np.array([1.5e-3, 0.8])
        num_lpdf = pdf.logpdf(x)
        ana_lpdf = log(1/sqrt(4*pi**2*np.linalg.det(cov))*exp(-np.dot(np.dot(x-c,np.linalg.inv(cov)),x-c)/2))
        self.assertAlmostEqual(num_lpdf, ana_lpdf, delta=1e-6)



import flavio
import numpy as np
from flavio.classes import AuxiliaryQuantity, Implementation
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.bdecays.wilsoncoefficients import wctot_dict
from flavio.physics.common import conjugate_par, conjugate_wc, add_dict
from flavio.config import config

# auxiliary function to construct transversity_amps_deltaC7, transversity_amps_deltaC9
def _transversity_amps_deltaC(q2, deltaC, C_name, par):
    scale = flavio.config['renormalization scale']['lambdab']
    mLb = par['m_Lambdab']
    mL = par['m_Lambda']
    mb = flavio.physics.running.running.get_mb(par, scale)
    ff = flavio.physics.bdecays.lambdablambdall.get_ff(q2, par)
    N = flavio.physics.bdecays.lambdablambdall.prefactor(q2, par, scale)
    ha = flavio.physics.bdecays.lambdablambdall.helicity_amps(q2, mLb, mL, ff)
    wc  = {'7': 0, '7p': 0, 'v': 0, 'a': 0, 's': 0, 'p': 0, 't': 0,'vp': 0, 'ap': 0, 'sp': 0, 'pp': 0, 'tp': 0, }
    wc[C_name] = deltaC
    return flavio.physics.bdecays.lambdablambdall.transverity_amps(ha, q2, mLb, mL, mb, 0, wc, N)


def transversity_amps_deltaC7(q2, deltaC7_dict, par):
    r"""A function returning a contribution to the transversity amplitudes in
    $\Lambda_b\to\Lambda\ell^+\ell^-$ coming from an effective transversity-dependent shift of
    the Wilson coefficient $C_7(\mu_b)$. This can be used to parametrize
    residual uncertainties due to subleading non-factorizable hadronic effects.

    The input dictionary `deltaC7_dict` should be of the form

    `{ 'perp0': deltaC7_perp0, 'para0': deltaC7_para0, 'perp1': deltaC7_perp1, 'para1': deltaC7_para1}`
    """
    ta = {}
    for amp in ['perp0', 'para0', 'perp1', 'para1']:
        for X in ['L', 'R']:
            ta[(amp, X)] = _transversity_amps_deltaC(q2, deltaC7_dict[amp], '7', par)[(amp, X)]
    return ta

def transversity_amps_deltaC9(q2, deltaC9_dict, par):
    r"""A function returning a contribution to the transversity amplitudes in
    $\Lambda_b\to\Lambda\ell^+\ell^-$ coming from an effective transversity-dependent shift of
    the Wilson coefficient $C_7(\mu_b)$. This can be used to parametrize
    residual uncertainties due to subleading non-factorizable hadronic effects.

    The input dictionary `deltaC9_dict` should be of the form

    `{ 'perp0': deltaC9_perp0, 'para0': deltaC9_para0, 'perp1': deltaC9_perp1, 'para1': deltaC9_para1}`
    """
    ta = {}
    for amp in ['perp0', 'para0', 'perp1', 'para1']:
        for X in ['L', 'R']:
            ta[(amp, X)] = _transversity_amps_deltaC(q2, deltaC9_dict[amp], 'v', par)[(amp, X)]
    return ta

# One possibility is to parametrize the effective shift in C7 or C9 as a simple
# polynomial in q2.

def transversity_amps_deltaC7_polynomial(q2, par):
    deltaC7_dict = {}
    for amp in ['perp0', 'para0', 'perp1', 'para1']:
        deltaC7_dict[amp]  = ( par['Lambdab->Lambda deltaC7 a_' + amp + ' Re']
                             + par['Lambdab->Lambda deltaC7 b_' + amp + ' Re'] *q2
                             + 1j*par['Lambdab->Lambda deltaC7 a_' + amp + ' Im']
                             + 1j*par['Lambdab->Lambda deltaC7 b_' + amp + ' Im'] *q2)
    return transversity_amps_deltaC7(q2, deltaC7_dict, par)

# a constant shift, e.g. for high q^2
def transversity_amps_deltaC9_constant(q2, par):
    deltaC9_dict = {}
    for amp in ['perp0', 'para0', 'perp1', 'para1']:
        deltaC9_dict[amp]  = ( par['Lambdab->Lambda deltaC9 c_' + amp + ' Re']
                             + 1j*par['Lambdab->Lambda deltaC9 c_' + amp + ' Im'])
    return transversity_amps_deltaC9(q2, deltaC9_dict, par)


def fct_deltaC7_polynomial(wc_obj, par_dict, q2, cp_conjugate):
    par = par_dict.copy()
    if cp_conjugate:
        par = conjugate_par(par)
    return transversity_amps_deltaC7_polynomial(q2, par)


def fct_deltaC9_constant(wc_obj, par_dict, q2, cp_conjugate):
    par = par_dict.copy()
    if cp_conjugate:
        par = conjugate_par(par)
    return transversity_amps_deltaC9_constant(q2, par_dict)


# AuxiliaryQuantity & Implementatation: subleading effects at LOW q^2

quantity = 'Lambdab->Lambdall subleading effects at low q2'
a = AuxiliaryQuantity(name=quantity,
                      arguments=['q2', 'cp_conjugate'])
a.description = (r'Contribution to $\Lambda_b\to \Lambda \ell^+\ell^-$ transversity amplitudes from'
                 r' subleading hadronic effects (i.e. all effects not included'
                 r' elsewhere) at $q^2$ below the charmonium resonances')

# Implementation: C7-polynomial
iname = 'Lambdab->Lambdall deltaC7 polynomial'
i = Implementation(name=iname, quantity=quantity,
               function=fct_deltaC7_polynomial)
i.set_description(r"Effective shift in the Wilson coefficient $C_7(\mu_b)$"
                  r" as a first-order polynomial in $q^2$.")

# AuxiliaryQuantity & Implementatation: subleading effects at HIGH q^2

quantity = 'Lambdab->Lambdall subleading effects at high q2'
a = AuxiliaryQuantity(name=quantity, arguments=['q2', 'cp_conjugate'])
a.description = ('Contribution to $\Lambda_b\to \Lambda \ell^+\ell^-$ transversity amplitudes from'
                ' subleading hadronic effects (i.e. all effects not included'
                r'elsewhere) at $q^2$ above the charmonium resonances')

# Implementation: C9 constant shift
iname = 'Lambdab->Lambdall deltaC9 shift'
i = Implementation(name=iname, quantity=quantity,
               function=fct_deltaC9_constant)
i.set_description(r"Effective constant shift in the Wilson coefficient $C_9(\mu_b)$.")



r"""Standard Model Wilson coefficients for $\Delta B=1$ transitions as well
as tree-level $B$ decays"""

from math import sqrt,pi,log
import pkgutil
import numpy as np
from io import StringIO
import scipy.interpolate
from flavio.physics.running import running
from flavio.physics import ckm
from flavio.physics.common import add_dict
from flavio.config import config
from flavio.physics.bdecays.common import meson_quark
from flavio.physics.bdecays import matrixelements
import flavio

# SM Wilson coefficients at 120 GeV in the basis
# [ C_1, C_2, C_3, C_4, C_5, C_6,
# C_7^eff, C_8^eff,
# C_9, C_10,
# C_3^Q, C_4^Q, C_5^Q, C_6^Q,
# Cb ]
# where all operators are defined as in hep-ph/0512066 *except*
# C_9,10, which are defined with an additional alpha/4pi prefactor.
_wcsm_120 = np.zeros(34)
_wcsm_120[:15] = np.array([  1.99030910e-01,   1.00285703e+00,  -4.17672471e-04,
         2.00964137e-03,   5.20961618e-05,   9.65703651e-05,
        -1.98501275e-01,  -1.09453204e-01,   1.52918563e+00,
        -4.06926405e+00,   6.15944332e-03,   0.00000000e+00,
        -1.12876870e-03,   0.00000000e+00,  -3.24099235e-03])

# di->djnunu Wilson coefficient
def CL_SM(par):
    r"""SM Wilson coefficient for $d_i\to d_j\nu\bar\nu$ transitions.

    This is implemented as an approximate formula as a function of the top
    mass."""
    # EW NLO corrections arXiv:1009.0947
    scale = 120. # <- result has very little sensitivity to high matching scale
    mt = flavio.physics.running.running.get_mt(par, scale)
    s2w = par['s2w']
    Xt0_165 = 1.50546 # LO result for mt=165, scale=120
    Xt0 = Xt0_165 * (1 + 1.14064 * (mt/165. - 1)) # LO
    Xt1 = Xt0_165 * (-0.031435 - 0.139303 * (mt/165. - 1)) # QCD NLO
    # (4.3), (4.4) of 1009.0947: NLO EW
    XtEW = Xt0 * (1 - 1.11508 + 1.12316*1.15338**(mt/165.)-0.179454*(mt/165)) - 1
    XtEW = XtEW * 0.00062392534457616328 # <- alpha_em/4pi at 120 GeV
    Xt = Xt0 + Xt1 + XtEW
    return -Xt/s2w


def wctot_dict(wc_obj, sector, scale, par, nf_out=None):
    r"""Get a dictionary with the total (SM + new physics) values  of the
    $\Delta F=1$ Wilson coefficients at a given scale, given a
    WilsonCoefficients instance."""
    wc_np_dict = wc_obj.get_wc(sector, scale, par, nf_out=nf_out)
    wcsm_120 = _wcsm_120.copy()
    wc_sm = running.get_wilson(par, wcsm_120, wc_obj.rge_derivative[sector], 120., scale, nf_out=nf_out)
    # now here comes an ugly fix. If we have b->s transitions, we should take
    # into account the fact that C7' = C7*ms/mb, and the same for C8, which is
    # not completely negligible. To find out whether we have b->s, we look at
    # the "sector" string.
    if sector[:2] == 'bs':
        # go from the effective to the "non-effective" WCs
        yi = np.array([0, 0, -1/3., -4/9., -20/3., -80/9.])
        zi = np.array([0, 0, 1, -1/6., 20, -10/3.])
        c7 = wc_sm[6] - np.dot(yi, wc_sm[:6]) # c7 (not effective!)
        c8 = wc_sm[7] - np.dot(zi, wc_sm[:6]) # c8 (not effective!)
        eps_s = running.get_ms(par, scale)/running.get_mb(par, scale)
        c7p = eps_s * c7
        c8p = eps_s * c8
        # go back to the effective WCs
        wc_sm[21] = c7p + np.dot(yi, wc_sm[15:21]) # c7p_eff
        wc_sm[22] = c7p + np.dot(zi, wc_sm[15:21]) # c8p_eff
    wc_labels = wc_obj.coefficients[sector]
    wc_sm_dict =  dict(zip(wc_labels, wc_sm))
    return add_dict((wc_np_dict, wc_sm_dict))

def get_wceff(q2, wc, par, B, M, lep, scale):
    r"""Get a dictionary with the effective $\Delta F=1$ Wilson coefficients
    in the convention appropriate for the generalized angular distributions.
    """
    xi_u = ckm.xi('u',meson_quark[(B,M)])(par)
    xi_t = ckm.xi('t',meson_quark[(B,M)])(par)
    qiqj=meson_quark[(B,M)]
    Yq2 = matrixelements.Y(q2, wc, par, scale, qiqj) + (xi_u/xi_t)*matrixelements.Yu(q2, wc, par, scale, qiqj)
        #   b) NNLO Q1,2
    delta_C7 = matrixelements.delta_C7(par=par, wc=wc, q2=q2, scale=scale, qiqj=qiqj)
    delta_C9 = matrixelements.delta_C9(par=par, wc=wc, q2=q2, scale=scale, qiqj=qiqj)
    mb = running.get_mb(par, scale)
    ll = lep + lep
    c = {}
    c['7']  = wc['C7eff_'+qiqj]      + delta_C7
    c['7p'] = wc['C7effp_'+qiqj]
    c['v']  = wc['C9_'+qiqj+ll]      + delta_C9 + Yq2
    c['vp'] = wc['C9p_'+qiqj+ll]
    c['a']  = wc['C10_'+qiqj+ll]
    c['ap'] = wc['C10p_'+qiqj+ll]
    c['s']  = mb * wc['CS_'+qiqj+ll]
    c['sp'] = mb * wc['CSp_'+qiqj+ll]
    c['p']  = mb * wc['CP_'+qiqj+ll]
    c['pp'] = mb * wc['CPp_'+qiqj+ll]
    c['t']  = 0
    c['tp'] = 0
    return c

def get_wceff_lfv(q2, wc, par, B, M, l1, l2, scale):
    r"""Get a dictionary with the effective $\Delta F=1$ Wilson coefficients
    with lepton flavour violation
    in the convention appropriate for the generalized angular distributions.
    """
    mb = running.get_mb(par, scale)
    qiqj=meson_quark[(B,M)]
    c = {}
    c['7']  = 0
    c['7p'] = 0
    c['v']  = wc['C9_'+qiqj+l1+l2]
    c['vp'] = wc['C9p_'+qiqj+l1+l2]
    c['a']  = wc['C10_'+qiqj+l1+l2]
    c['ap'] = wc['C10p_'+qiqj+l1+l2]
    c['s']  = mb * wc['CS_'+qiqj+l1+l2]
    c['sp'] = mb * wc['CSp_'+qiqj+l1+l2]
    c['p']  = mb * wc['CP_'+qiqj+l1+l2]
    c['pp'] = mb * wc['CPp_'+qiqj+l1+l2]
    c['t']  = 0
    c['tp'] = 0
    return c


def get_wceff_nunu(q2, wc, par, B, M, nu1, nu2, scale):
    r"""Get a dictionary with the effective $\Delta F=1$ Wilson coefficients
    with neutrinos in the final state
    in the convention appropriate for the generalized angular distributions.
    """
    qiqj=meson_quark[(B,M)]
    c = {}
    c['7']  = 0
    c['7p'] = 0
    c['v']  =  wc['CL_'+qiqj+nu1+nu2]
    c['vp'] =  wc['CR_'+qiqj+nu1+nu2]
    c['a']  = -wc['CL_'+qiqj+nu1+nu2]
    c['ap'] = -wc['CR_'+qiqj+nu1+nu2]
    c['s']  = 0
    c['sp'] = 0
    c['p']  = 0
    c['pp'] = 0
    c['t']  = 0
    c['tp'] = 0
    return c

def get_CVSM(par, scale, nf):
    r"""Get the Wilson coefficient of the operator $C_V$ in $d_i\to d_j\ell\nu$
    in the SM including EW corrections."""
    if nf >= 4: # for B and D physics
        alpha_e = running.get_alpha(par, scale)['alpha_e']
        return 1 + alpha_e/pi * log(par['m_Z']/scale)
    else: # for K and pi physics
        # Marciano & Sirlin 1993
        return sqrt(1.0232)

def get_wceff_fccc(wc_obj, par, qiqj, lep, mqi, scale, nf=5):
    r"""Get a dictionary with the $d_i\to d_j$ Wilson coefficients
    in the convention appropriate for the generalized angular distributions.
    """
    qqlnu = qiqj + lep + 'nu'
    wc = wc_obj.get_wc(qqlnu, scale, par)
    c_sm = get_CVSM(par, scale, nf)
    c = {}
    c['7']  = 0
    c['7p'] = 0
    c['v']  = (c_sm + wc['CV_'+qqlnu])/2.
    c['vp'] = wc['CVp_'+qqlnu]/2.
    c['a']  = -(c_sm + wc['CV_'+qqlnu])/2.
    c['ap'] = -wc['CVp_'+qqlnu]/2.
    c['s']  = 1/2 * mqi * wc['CS_'+qqlnu]/2.
    c['sp'] = 1/2 * mqi * wc['CSp_'+qqlnu]/2.
    c['p']  = -1/2 * mqi * wc['CS_'+qqlnu]/2.
    c['pp'] = -1/2 * mqi * wc['CSp_'+qqlnu]/2.
    c['t']  = wc['CT_'+qqlnu]
    c['tp'] = 0
    return c



"""Main functions for user interaction. All of these are imported into the
top-level namespace."""

import flavio
import numpy as np
from collections import OrderedDict

def np_prediction(obs_name, wc_obj, *args, **kwargs):
    """Get the central value of the new physics prediction of an observable.

    Parameters
    ----------

    - `obs_name`: name of the observable as a string
    - `wc_obj`: an instance of `flavio.WilsonCoefficients`

    Additional arguments are passed to the observable and are necessary,
    depending on the observable (e.g. $q^2$-dependent observables).
    """
    obs = flavio.classes.Observable.get_instance(obs_name)
    return obs.prediction_central(flavio.default_parameters, wc_obj, *args, **kwargs)

def sm_prediction(obs_name, *args, **kwargs):
    """Get the central value of the Standard Model prediction of an observable.

    Parameters
    ----------

    - `obs_name`: name of the observable as a string

    Additional arguments are passed to the observable and are necessary,
    depending on the observable (e.g. $q^2$-dependent observables).
    """
    obs = flavio.classes.Observable.get_instance(obs_name)
    wc_sm = flavio.WilsonCoefficients()
    return obs.prediction_central(flavio.default_parameters, wc_sm, *args, **kwargs)

def sm_uncertainty(obs_name, *args, N=100, **kwargs):
    """Get the uncertainty of the Standard Model prediction of an observable.

    Parameters
    ----------

    - `obs_name`: name of the observable as a string
    - `N` (optional): number of random evaluations of the observable.
    The relative accuracy of the uncertainty returned is given by $1/\sqrt{2N}$.

    Additional arguments are passed to the observable and are necessary,
    depending on the observable (e.g. $q^2$-dependent observables).
    """
    obs = flavio.classes.Observable.get_instance(obs_name)
    wc_sm = flavio.WilsonCoefficients()
    par_random = [flavio.default_parameters.get_random_all() for i in range(N)]
    all_pred = np.array([
        obs.prediction_par(par, wc_sm, *args, **kwargs)
        for par in par_random
    ])
    return np.std(all_pred)

def sm_error_budget(obs_name, *args, N=50, **kwargs):
    """Get the *relative* uncertainty of the Standard Model prediction due to
    variation of individual observables.

    Parameters
    ----------

    - `obs_name`: name of the observable as a string
    - `N` (optional): number of random evaluations of the observable.
    The relative accuracy of the uncertainties returned is given by $1/\sqrt{2N}$.

    Additional arguments are passed to the observable and are necessary,
    depending on the observable (e.g. $q^2$-dependent observables).
    """
    obs = flavio.classes.Observable.get_instance(obs_name)
    wc_sm = flavio.WilsonCoefficients()
    par_central = flavio.default_parameters.get_central_all()
    par_random = [flavio.default_parameters.get_random_all() for i in range(N)]

    # Step 1: determine the parameters the observable depends on at all.
    # to this end, compute the observables once for each parameter with a
    # random value for this parameter but central values for all other
    # parameters. If the prediction is equal to the central prediction, the
    # observable does not depend on the parameter!
    pred_central = obs.prediction_par(par_central, wc_sm, *args, **kwargs)
    dependent_par = []
    for k in par_central.keys():
        par_tmp = par_central.copy()
        par_tmp[k] = par_random[0][k]
        pred_tmp = obs.prediction_par(par_tmp, wc_sm, *args, **kwargs)
        if pred_tmp != pred_central:
            dependent_par.append(k)

    # Step 2: for each of the dependent parameters, determine the error
    # analogous to the sm_uncertainty function. Normalize to the central
    # prediction (so relative errors are returned)
    individual_errors = {}
    def make_par_random(key, par_random):
        par_tmp = par_central.copy()
        par_tmp[key] = par_random[key]
        return par_tmp
    for p in dependent_par:
        par_random_p = [make_par_random(p, pr) for pr in par_random]
        all_pred = np.array([
            obs.prediction_par(par, wc_sm, *args, **kwargs)
            for par in par_random_p
        ])
        individual_errors[p] = np.std(all_pred)/abs(pred_central)
    return individual_errors

def sm_covariance(obs_list, N=100, par_vary='all', **kwargs):
    """Get the covariance matrix of the Standard Model predictions for a
    list of observables.

    Parameters
    ----------

    - `obs_list`: a list of observables that should be given either as a string
    name (for observables that do not depend on any arguments) or as a tuple
    of a string and values for the arguements the observable depends on (e.g.
    the values of `q2min` and `q2max` for a binned observable)
    - `N` (optional): number of random evaluations of the observables.
    The relative accuracy of the uncertainties returned is given by $1/\sqrt{2N}$.
    - `par_vary`: a list of parameters to vary. Defaults to 'all', i.e. all
    parameters are varied according to their probability distributions.
    """
    wc_sm = flavio.WilsonCoefficients()
    par_central_all = flavio.default_parameters.get_central_all()
    par_random_all = [flavio.default_parameters.get_random_all() for i in range(N)]
    def par_random_some(par_random, par_central):
        # take the central values for the parameters not to be varied
        par1 = {k: v for k, v in par_central.items() if k not in par_vary}
        # take the random values for the parameters to be varied
        par2 = {k: v for k, v in par_random.items() if k in par_vary}
        par1.update(par2) # merge them
        return par1
    if par_vary == 'all':
        par_random = par_random_all
    else:
        par_random = [par_random_some(par_random_all[i], par_central_all) for i in range(N)]
    def get_prediction(obs, par):
        if isinstance(obs, str):
             obs_obj = flavio.classes.Observable.get_instance(obs)
             return obs_obj.prediction_par(par, wc_sm, **kwargs)
        elif isinstance(obs, tuple):
             obs_obj = flavio.classes.Observable.get_instance(obs[0])
             return obs_obj.prediction_par(par, wc_sm, *obs[1:], **kwargs)
    all_pred = np.array([
        [get_prediction(obs, par)
            for par in par_random
        ]
        for obs in obs_list
    ])
    return np.cov(all_pred)



"""Functions for exclusive $B_s\to V\ell^+\ell^-$ decays, taking into account
the finite life-time difference between the $B_s$ mass eigenstates,
see arXiv:1502.05509."""

import flavio
from . import observables
from flavio.classes import Observable, Prediction
import cmath

def bsvll_obs(function, q2, wc_obj, par, B, V, lep):
    ml = par['m_'+lep]
    mB = par['m_'+B]
    mV = par['m_'+V]
    y = par['DeltaGamma/Gamma_'+B]/2.
    if q2 < 4*ml**2 or q2 > (mB-mV)**2:
        return 0
    scale = flavio.config['renormalization scale']['bvll']
    mb = flavio.physics.running.running.get_mb(par, scale)
    ff = flavio.physics.bdecays.bvll.amplitudes.get_ff(q2, par, B, V)
    h = flavio.physics.bdecays.bvll.amplitudes.helicity_amps(q2, wc_obj, par, B, V, lep)
    h_bar = flavio.physics.bdecays.bvll.amplitudes.helicity_amps_bar(q2, wc_obj, par, B, V, lep)
    J = flavio.physics.bdecays.angular.angularcoeffs_general_v(h, q2, mB, mV, mb, 0, ml, ml)
    J_bar = flavio.physics.bdecays.angular.angularcoeffs_general_v(h_bar, q2, mB, mV, mb, 0, ml, ml)
    h_tilde = h_bar.copy()
    h_tilde[('pl', 'V')] = h_bar[('mi', 'V')]
    h_tilde[('pl', 'A')] = h_bar[('mi', 'A')]
    h_tilde[('mi', 'V')] = h_bar[('pl', 'V')]
    h_tilde[('mi', 'A')] = h_bar[('pl', 'A')]
    h_tilde['S'] = -h_bar['S']
    q_over_p = flavio.physics.mesonmixing.observables.q_over_p(wc_obj, par, B)
    phi = cmath.phase(-q_over_p) # the phase of -q/p
    J_h = flavio.physics.bdecays.angular.angularcoeffs_h_v(phi, h, h_tilde, q2, mB, mV, mb, 0, ml, ml)
    return function(y, J, J_bar, J_h)

def S_theory_num_Bs(y, J, J_bar, J_h, i):
    # (42) of 1502.05509
    return 1/(1-y**2) * (J[i] + J_bar[i]) - y/(1-y**2) * J_h[i]

def S_experiment_num_Bs(y, J, J_bar, J_h, i):
    if i in [4, '6s', '6c', 7, 9]:
        return -S_theory_num_Bs(y, J, J_bar, J_h, i)
    return S_theory_num_Bs(y, J, J_bar, J_h, i)


def S_experiment_Bs(y, J, J_bar, J_h, i):
    r"""CP-averaged angular observable $S_i$ in the LHCb convention.

    See eq. (C.8) of arXiv:1506.03970v2.
    """
    return S_experiment_num_Bs(y, J, J_bar, J_h, i)/SA_den_Bs(y, J, J_bar, J_h)

def dGdq2_ave_Bs(y, J, J_bar, J_h):
    # (48) of 1502.05509
    return (1/(1-y**2) * (observables.dGdq2(J) + observables.dGdq2(J_bar))
            - y/(1-y**2) * observables.dGdq2(J_h))/2.

# denominator of S_i and A_i observables
def SA_den_Bs(y, J, J_bar, J_h):
    return 2*dGdq2_ave_Bs(y, J, J_bar, J_h)

def FL_Bs(y, J, J_bar, J_h):
    r"""Longitudinal polarization fraction $F_L$"""
    return FL_num_Bs(y, J, J_bar, J_h)/SA_den_Bs(y, J, J_bar, J_h)

def FL_num_Bs(y, J, J_bar, J_h):
    return -S_theory_num_Bs(y, J, J_bar, J_h, '2c')


def bsvll_dbrdq2(q2, wc_obj, par, B, V, lep):
    tauB = par['tau_'+B]
    return tauB * bsvll_obs(dGdq2_ave_Bs, q2, wc_obj, par, B, V, lep)

def bsvll_obs_int(function, q2min, q2max, wc_obj, par, B, V, lep):
    def obs(q2):
        return bsvll_obs(function, q2, wc_obj, par, B, V, lep)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)

def bsvll_dbrdq2_int(q2min, q2max, wc_obj, par, B, V, lep):
    def obs(q2):
        return bsvll_dbrdq2(q2, wc_obj, par, B, V, lep)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)/(q2max-q2min)

# Functions returning functions needed for Prediction instances

def bsvll_dbrdq2_int_func(B, V, lep):
    def fct(wc_obj, par, q2min, q2max):
        return bsvll_dbrdq2_int(q2min, q2max, wc_obj, par, B, V, lep)
    return fct

def bsvll_dbrdq2_func(B, V, lep):
    def fct(wc_obj, par, q2):
        return bsvll_dbrdq2(q2, wc_obj, par, B, V, lep)
    return fct

def bsvll_obs_int_ratio_func(func_num, func_den, B, V, lep):
    def fct(wc_obj, par, q2min, q2max):
        num = bsvll_obs_int(func_num, q2min, q2max, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom = bsvll_obs_int(func_den, q2min, q2max, wc_obj, par, B, V, lep)
        return num/denom
    return fct

def bsvll_obs_int_ratio_leptonflavour(func, B, V, l1, l2):
    def fct(wc_obj, par, q2min, q2max):
        num = bsvll_obs_int(func, q2min, q2max, wc_obj, par, B, V, l1)
        if num == 0:
            return 0
        denom = bsvll_obs_int(func, q2min, q2max, wc_obj, par, B, V, l2)
        return num/denom
    return fct

def bsvll_obs_ratio_func(func_num, func_den, B, V, lep):
    def fct(wc_obj, par, q2):
        num = bsvll_obs(func_num, q2, wc_obj, par, B, V, lep)
        if num == 0:
            return 0
        denom = bsvll_obs(func_den, q2, wc_obj, par, B, V, lep)
        return num/denom
    return fct

# Observable and Prediction instances

_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_observables = {
'FL': {'func_num': FL_num_Bs, 'tex': r'\overline{F_L}', 'desc': 'Time-averaged longitudinal polarization fraction'},
'S3': {'func_num': lambda y, J, J_bar, J_h: S_experiment_num_Bs(y, J, J_bar, J_h, 3), 'tex': r'\overline{S_3}', 'desc': 'Time-averaged, CP-averaged angular observable'},
'S4': {'func_num': lambda y, J, J_bar, J_h: S_experiment_num_Bs(y, J, J_bar, J_h, 4), 'tex': r'\overline{S_4}', 'desc': 'Time-averaged, CP-averaged angular observable'},
'S7': {'func_num': lambda y, J, J_bar, J_h: S_experiment_num_Bs(y, J, J_bar, J_h, 7), 'tex': r'\overline{S_7}', 'desc': 'Time-averaged, CP-averaged angular observable'},
}
_hadr = {
'Bs->phi': {'tex': r"B_s\to \phi ", 'B': 'Bs', 'V': 'phi', },
}
for l in ['e', 'mu', 'tau']:
    for M in _hadr.keys():
        for obs in sorted(_observables.keys()):

            # binned angular observables
            _obs_name = "<" + obs + ">("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description('Binned ' + _observables[obs]['desc'] + r" in $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
            _obs.tex = r"$\langle " + _observables[obs]['tex'] + r"\rangle(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
            Prediction(_obs_name, bsvll_obs_int_ratio_func(_observables[obs]['func_num'], SA_den_Bs, _hadr[M]['B'], _hadr[M]['V'], l))

            # differential angular observables
            _obs_name = obs + "("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2'])
            _obs.set_description(_observables[obs]['desc'][0].capitalize() + _observables[obs]['desc'][1:] + r" in $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
            _obs.tex = r"$" + _observables[obs]['tex'] + r"(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
            Prediction(_obs_name, bsvll_obs_ratio_func(_observables[obs]['func_num'], SA_den_Bs, _hadr[M]['B'], _hadr[M]['V'], l))

        # binned branching ratio
        _obs_name = "<dBR/dq2>("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description(r"Binned time-integrated differential branching ratio of $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
        _obs.tex = r"$\langle \frac{d\overline{\text{BR}}}{dq^2} \rangle(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
        Prediction(_obs_name, bsvll_dbrdq2_int_func(_hadr[M]['B'], _hadr[M]['V'], l))

        # differential branching ratio
        _obs_name = "dBR/dq2("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(r"Differential time-integrated branching ratio of $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
        _obs.tex = r"$\frac{d\overline{\text{BR}}}{dq^2}(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
        Prediction(_obs_name, bsvll_dbrdq2_func(_hadr[M]['B'], _hadr[M]['V'], l))

# Lepton flavour ratios
for l in [('mu','e'), ('tau','mu'),]:
    for M in _hadr.keys():

            # binned ratio of BRs
            _obs_name = "<R"+l[0]+l[1]+">("+M+"ll)"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description(r"Ratio of partial branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
            _obs.tex = r"$\langle R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} \rangle(" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
            Prediction(_obs_name, bsvll_obs_int_ratio_leptonflavour(dGdq2_ave_Bs, _hadr[M]['B'], _hadr[M]['V'], *l))



import unittest
import numpy as np
import flavio
from .fits import *
from flavio.classes import *
from flavio.statistics.probability import *
from flavio.config import config
import copy

class TestClasses(unittest.TestCase):
    def test_fit_class(self):
        o = Observable( 'test_obs' )
        d = NormalDistribution(4.2, 0.2)
        m = Measurement( 'measurement of test_obs' )
        m.add_constraint(['test_obs'], d)
        par = copy.deepcopy(flavio.parameters.default_parameters)
        par.set_constraint('m_b', '4.2+-0.2')
        par.set_constraint('m_c', '1.2+-0.1')
        with self.assertRaises(AssertionError):
            # same parameter as fit and nuisance
            Fit('test_fit_1', par, ['m_b'], ['m_b'], ['test_obs'])
        with self.assertRaises(AssertionError):
            # non-existent fit parameter
            Fit('test_fit_1', par, ['blabla'],  [], ['test_obs'])
        with self.assertRaises(AssertionError):
            # non-existent nuisance parameter
            Fit('test_fit_1', par, [],  ['blabla'], ['test_obs'])
        def wc_fct(C):
            return {'CVLL_bsbs': C}
        fit = Fit('test_fit_1', par, ['m_b'],  ['m_c'], ['test_obs'], ['C'], wc_fct)
        self.assertEqual(fit.fit_parameters, ['m_b'])
        self.assertEqual(fit.nuisance_parameters, ['m_c'])
        self.assertEqual(fit.fit_wc_names, ['C'])
        self.assertEqual(fit.get_measurements, ['measurement of test_obs'])
        self.assertEqual(fit.get_central_fit_parameters, [4.2])
        self.assertEqual(fit.get_central_nuisance_parameters, [1.2])
        # removing dummy instances
        Fit.del_instance('test_fit_1')
        Observable.del_instance('test_obs')
        Measurement.del_instance('measurement of test_obs')

    def test_bayesian_fit_class(self):
        o = Observable( 'test_obs 2' )
        o.arguments = ['q2']
        def f(wc_obj, par_dict, q2):
            return par_dict['m_b']*2 *q2
        pr  = Prediction( 'test_obs 2', f )
        d = NormalDistribution(4.2, 0.2)
        m = Measurement( 'measurement of test_obs 2' )
        m.add_constraint([('test_obs 2', 3)], d)
        par = copy.deepcopy(flavio.parameters.default_parameters)
        par.set_constraint('m_b', '4.2+-0.2')
        par.set_constraint('m_c', '1.2+-0.1')
        par.set_constraint('m_s', '0.10(5)')
        wc = flavio.physics.eft.WilsonCoefficients()
        def wc_fct(CL, CR):
            return {'CVLL_bsbs': CL, 'CVRR_bsbs': CR}
        fit = BayesianFit('bayesian_test_fit_1', par, ['m_b','m_c'],  ['m_s'], [('test_obs 2',3)], ['CL','CR'], wc_fct)
        self.assertEqual(fit.get_measurements, ['measurement of test_obs 2'])
        self.assertEqual(fit.dimension, 5)
        # test from array to dict ...
        d = fit.array_to_dict(np.array([1.,2.,3.,4.,5.]))
        self.assertEqual(d, {'nuisance_parameters': {'m_s': 3.0}, 'fit_parameters': {'m_c': 2.0, 'm_b': 1.0}, 'fit_wc': {'CL': 4.0, 'CR': 5.0}})
        # ... and back
        np.testing.assert_array_equal(fit.dict_to_array(d), np.array([1.,2.,3.,4.,5.]))
        self.assertEqual(fit.get_random.shape, (5,))
        fit.log_prior_parameters(np.array([4.5,1.0,0.08,4.,5.]))
        fit.get_predictions(np.array([4.5,1.0,0.08,4.,5.]))
        fit.log_likelihood(np.array([4.5,1.0,0.08,4.,5.]))
        # removing dummy instances
        BayesianFit.del_instance('bayesian_test_fit_1')
        Observable.del_instance('test_obs 2')
        Measurement.del_instance('measurement of test_obs 2')



"""Module for parsing SLHA-like files. Adapted from the Rosetta package,
https://rosetta.hepforge.org/
http://arxiv.org/abs/1508.05895
"""

from collections import OrderedDict, MutableMapping
from operator import itemgetter
import sys
import re
import os
import logging

log = logging.getLogger('SLHA')

class CaseInsensitiveDict(MutableMapping):
    '''
    Dict class for string keys that behaves in a case insensitive way.
    '''
    def __init__(self, data=None, **kwargs):
        self._data = {}
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._data[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._data[key.lower()][1]

    def __delitem__(self, key):
        del self._data[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in list(self._data.values()))

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return repr([(k,v) for (k,v) in list(self._data.values())])

class CaseInsensitiveOrderedDict(MutableMapping):
    '''
    OrderedDict class for string keys that behaves in a case insensitive way.
    '''
    def __init__(self, data=None, **kwargs):
        self._data = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def __setitem__(self, key, value):
        self._data[key.lower()] = (key, value)

    def __getitem__(self, key):
        return self._data[key.lower()][1]

    def __delitem__(self, key):
        del self._data[key.lower()]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in list(self._data.values()))

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return repr([(k,v) for (k,v) in list(self._data.values())])

class Block(MutableMapping):
    '''
    Container class for SLHA block with a single counter. A subclass of
    `collections.OrderedDict`with a restriction on integer keys and a modified
    __repr__(). Block can't be initialised with any positional arguments but
    rather with the 'data' keyword argument. It can optionally be named using
    the 'name' keyword argument. The__str__ function are also defined to output
    the contained data in an SLHA formatted block.
    '''

    def __checkkey__(self, key):
        '''Forces key to be of integer type.'''
        if type(key) is not self.keytype:
            raise TypeError("Key: '{}'. ".format(key) + self.__class__.__name__
                            + ' only accepts keys of {}.'.format(self.keytype))
        else:
            return key

    def __cast__(self, val):
        '''
        Attempts to cast values to type specified in 'vtype' keyword argument
        of constructor.
        '''
        try:
            return self.cast(val)
        except ValueError:
            return val

    def __init__(self, name=None, decimal=5, ktype=int, vtype=float,
                       preamble='', scale=None, data=None):
        '''    Intialisation keyword arguments:
            name - A name for the block that will appear
                   in the __str__ and __repr__ methods.
            scale - a renormalization scale in GeV
            data - A dict type object supporting iteritems()
                   to initialize Block values.
            decimal - Number of decimal points with which
                      to write out parameter values.
            vtype - a function to cast parameter values if
                    read in as strings i.e. int, float.
            preamble - Some text to print before the __str__ output.
        '''
        self._data = OrderedDict()
        self.name = name
        self.scale = scale
        self.keytype = ktype
        self.cast = vtype # value casting function
        self.fmt = ':+.{}e'.format(decimal) # format string
        self.preamble = preamble
        if data is not None:
            self._data.update(data)

    def __setitem__(self, key,value):
        self._data[self.__checkkey__(key)] = self.__cast__(value)

    def __getitem__(self, key):
        return self._data[self.__checkkey__(key)]

    def __contains__(self, key):
        return key in self._data

    def __delitem__(self, key):
        del self._data[self.__checkkey__(key)]

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return ('<{}: "{}"; {} entries.>'.format(self.__class__,
                                                      self.name, len(self)))
    def __str__(self):
        content = []
        sortitems = sorted(self.items())
        for k,v in sortitems:
            try:
                val = float(v)
                fmt = self.fmt
            except ValueError:
                val = v
                fmt = ''

            line = ('    {{: <4}} {{{}}}\n'.format(fmt)).format
            content.append(line(k,val))
        string = self.preamble+'\n'
        if content:
            string += 'BLOCK {}\n'.format(self.name) + ''.join(content)
        if self.scale is not None:
            string += ' Q=' + str(scale)
        return string

    def dict(self):
        '''Return SHLA Block data as a regular python dict.'''
        return self._data


class Matrix(Block):
    '''
    Container class for SLHA block with multiple counters. A subclass of
    `collections.OrderedDict`with a restriction on tuple keys and a modified
    __repr__(). Block can't be initialised with any positional arguments but
    rather with the 'data' keyword argument. It can optionally be named using
    the 'name' keyword argument. The__str__ function are also defined to output
    the contained data in an SLHA formatted block. The object is indexed
    like a numpy multi dimensional array ( x[i,j,k,...] ).
    '''

    def __init__(self, *args, **kwargs):
        kwargs['ktype'] = tuple
        super(Matrix, self).__init__(*args, **kwargs)

    def __str__(self):
        content = []
        sortitems = sorted(list(self.items()), key=itemgetter(0,0))
        for k,v in sortitems:
            try:
                v = float(v)
                fmt = self.fmt
            except ValueError:
                fmt=''
            ind = '{: <4}'*len(k)
            line = ('    {} {{{}}}\n'.format(ind,fmt)).format
            args = list(k)+[v]
            content.append(line(*args))

        string = self.preamble+'\n'

        if content:
            string += 'BLOCK {}\n'.format(self.name) + ''.join(content)

        return string

    def dimension(self):
        indx = []
        for i,k in enumerate(self.keys()):
            if i==0:
                indx = [[] for _ in k]
            for j, l in enumerate(k):
                indx[j].append(l)

        isets = [sorted(list(set(x))) for x in indx]

        for i,s in enumerate(isets):
            smin, smax = min(s), max(s)

            if not s == list(range(smin,smax+1)):
                strs = [str(x) for x in s]
                err = ('subdimension {} of array '.format(i) +
                       '{}, ({}), is not complete range.'.format(repr(self),
                                                                ','.join(strs)))
                raise SLHAError(err)

        return tuple( [len(x) for x in isets] )

    def array(self):
        array = []
        for k,v in self.items():
            if isinstance(v,Matrix):
                array.append(v.array())
            else:
                # array.append([v for v in v.values()])
                array.append(v)
        return array


class CBlock(Block):
    container = Block
    def __init__(self, *args):
        if not args:
            re, im = self.container(name='_re'), self.container(name='_im')
            self.__init__(re,im)
            return
        else:
            try:
                real, imag = args
            except ValueError:
                err = ('{}.__init__() takes 0 '.format(self.__class__)+
                       'or 2 arguments ({} given)'.format(len(args)))
                raise TypeError(err)
        assert type(real) is self.container and type(imag) is self.container, (
                ' Cblock constructor takes 2 {} '.format(self.container)
                + 'objects as arguments')
        name, decimal, preamble, scale = real.name, int(real.fmt[-2]), real.preamble, real.scale
        super(CBlock, self).__init__(name=name, decimal=decimal,
                                     preamble=preamble, scale=scale, vtype=complex)
        self.fill(real, imag)

    def fill(self, real, imag):
        '''
        Fill a CBlock from two Block instances containing the
        real and imaginary parts
        '''
        self._re, self._im = real, imag
        for k, re in real.items():
            try:
                im = imag[k]
                entry = complex(re,im)
            except KeyError:
                entry = re
            self[k] = entry

    def __setitem__(self, key, value):
        cval = complex(value)
        re, im = value.real, value.imag
        self._re[key], self._im[key] = re, im
        super(CBlock, self).__setitem__(key, cval)

    def __str__(self):
        string = self.preamble
        if self._re._data:
            string += str(self._re)
        if self._im._data:
            string += str(self._im)
        return string

class CMatrix(CBlock, Matrix):
    container = Matrix


class Decay(MutableMapping):
    '''
    Container class for SLHA Decay blocks. A subclass of `collections.OrderedDict`
    with a restriction on tuple keys and float values less than 1. A modified
    __repr__() function is implemented for easy writing to file. Decay is
    initialised with a PID argument to specify the particle to which it refers
    as well as its total width. The 'sum of' branching ratios is kept and a
    ValueError will be raised if the total exceeds 1. It can optionally be
    named using the 'name' keyword argument. Finally a __str__() function is
    also defined to output the SLHA formatted block.
    '''

    def __checkkey__(self, key):
        '''    Forces key to be a tuple and casts the elements to integers.'''
        if type(key) is not tuple:
            raise TypeError( self.__class__.__name__ +
                             ' only accepts tuple keys: (PID1, PID2,...).' )
        else:
            return tuple(map(int,key))

    def __checkval__(self, val):
        '''    Forces values (i.e. branching ratios) to be a float less than 1.'''
        try:
            fval = float(val)
        except ValueError:
            raise TypeError( self.__class__.__name__ +
                         ' only accepts floats or values castable via float()' )
        if fval > 1.:
            raise TypeError("SLHA Decay object for PID = {}. ".format(self.PID)+
                            "Branching ratio > 1 encountered : {}".format(fval))
        return fval

    def __init__(self, PID, total,
                 data=None, comment='', decimal=5, preamble= ''):
        '''    Positional arguments:
            PID - integer to denote particle whose decay is being described
            total - total width

            Keyword arguments:
            data - A dict type object supporting iteritems()
                   to initialize Decay values.
            comment - prints a comment to the right of the block
                      declaration in __str__().
            decimal - Number of decimal points with which
                      to write out width and BRs.
            preamble - Some text to print before the __str__ output.'''

        try:
            self.PID=int(PID)
        except ValueError:
            err = ("SLHA Decay object for PID = {}. ".format(PID) +
                   "'PID' argument must be an integer or be castable via int()")
            raise TypeError(err)

        try:
            self.total=float(total)
        except ValueError:
            err = ("SLHA Decay object for PID = {}. ".format(PID) +
                  "'total' argument must be a float or be castable via float()")
            raise TypeError(err)

        self._BRtot = 0.
        self._fmt = ':+.{}e'.format(decimal)
        self._comment = comment
        self._decimal = decimal
        self._comments = {}
        self.preamble = preamble
        self._data = OrderedDict()
        if data is not None:
            self._data.update(data)


    def __setitem__(self, key, value):
        self._data[self.__checkkey__(key)] = self.__checkval__(value)
        self._BRtot+=self.__checkval__(value)
        # if self._BRtot > 1.:
        #     log.error('!!! ERROR !!!')
        #     log.error(self)
        #     raise ValueError("SLHA Decay object for PID = {}. ".format(self.PID)
        #                     + "Sum of branching ratios > 1!")

    def __getitem__(self, key):
        return self._data[self.__checkkey__(key)]

    def __delitem__(self, key):
        del self._data[self.__checkkey__( key )]
        self._BRtot-=self.__checkval__(self[key])

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return ( '<SHLA Decay: PID={}; {} entries.>'.format(self.PID,
                                                            len(self)) )

    def __str__(self):
        above = '#        PDG        Width\n'
        title = ('DECAY    {{:<10}} {{{}}} # {{}}\n'.format(self._fmt)).format
        below = '#    BR{}    NDA       ID1       ID2...\n'.format(' '*self._decimal)
        if len(self)==0: below=''
        content = []
        for k,v in self.items():
            nparts = len(k)
            idfmt = nparts*'{: <10}'
            line = ('{{{}}}    {{: <10}}{}'.format(self._fmt,idfmt)).format

            cmnt = ('# {}'.format(self._comments[k])
                     if k in self._comments else '')

            content.append(line(v, nparts, *k) + cmnt)

        string = (self.preamble + '\n' + above
                 + title(self.PID, self.total, self._comment)
                 + below + '\n'.join(content) )
        return string

    def new_channel(self, PIDs, BR, comment=None):
        '''
        Add a new decay channel.
        PIDs - tuple of integers.
        BR - the branching ratio into that channel.
        comment - optional comment to be written to the
                  right of that channel in __str__().
        '''
        self[PIDs]=BR
        if comment is not None: self._comments[PIDs]=comment

class Card(object):
    '''
    SLHA card object: a container for storing multiple SLHA.Block,
    SLHA.NamedBlock and SLHA.Decay objects. Blocks and decays are stored in
    OrderedDicts self.blocks and self.decays respectively.
    Index syntax mycard[key] can be used to get or set elements of blocks or
    decays possessed by the Card instance. If passed a string key, the card
    instance will look up the first block in self.blocks possessing a parameter
    with the given name while an int key will return the SLHA.Decay object for
    that PID.
    '''
    def _parent_block(self, key):
        '''
        Returns the parent of a given string key. First searches for a Matrix
        object with that name, then searches in self.matrices for a Matrix
        containing that name and finally searches in self.blocks for a block
        containing that name. Otherwise raises Key error.
        '''

        if self.has_matrix(key): return self.matrices

        for block in list(self.matrices.values()):
            if key in block: return block
            try:
                if key in block._re:
                    return block._re
                if key in block._im:
                    return block._im
            except AttributeError:
                pass

        for block in list(self.blocks.values()):
            if key in block: return block

        return None

    def _container(self, key):
        '''
        Returns the container containing the indexed key. Used in overloading
        __setitem__, __getitem__, __contains__ and __delitem__.
        '''
        if type(key) is int:
            return self.decays
        elif type(key) is str:
            return self._parent_block(key.lower())
        else:
            err = ('SLHA Card has integer keys for '
                   'DECAY and string keys for BLOCK.')
            raise ValueError(err)

    def __init__(self, blocks=None, decays=None, name=None):
        '''    Keyword arguments:
        blocks - a dictionary of items with which to initialise
                 individual blocks in the card instance i.e.
                 name:block pairs with block an SLHA.NamedBlock.
        decays - a dictionary of SLHA.Decay objects.
        matrices - a dictionary of SLHA.NamedMatrix objects.
        name - card name'''

        self.blocks = CaseInsensitiveOrderedDict()
        self.decays = OrderedDict()
        self.matrices = CaseInsensitiveOrderedDict()
        self.name = name if name is not None else ''
        if blocks is not None:
            for bname, block in blocks.items():
                self.blocks[bname] = NamedBlock(name=bname, data=block)
        if decays is not None:
            for PID, decay in decays.items():
                self.decays[PID]= Decay(PID,decay.total, data=decay)

    def __repr__(self):
        return ('<SHLA Card "{}": {} blocks, {} decays.>'.format(self.name,
                       len(self.matrices)+ len(self.blocks), len(self.decays)))
    def __contains__(self, key):
        container = self._container(key)
        return True if container is not None else False

    def __getitem__(self, key):
        container = self._container(key)
        if container is not None:
            return container[key.lower()]
        else:
            err = 'Key/Matrix "{}" not found in {}'.format(key, self)
            raise KeyError(err)

    def __delitem__(self, key):
        container = self._container(key)
        if container is not None:
            del container[key.lower()]
        else:
            err = 'Key/Matrix "{}" not found in {}'.format(key, self)
            raise KeyError(err)

    def __setitem__(self, key, value):
        container = self._container(key)
        if container is not None:
            return container.__setitem__(key.lower(), value)
        else:
            err = 'Key/Matrix "{}" not found in {}'.format(key, self)
            raise KeyError(err)

    def add_block(self, block, preamble=''):
        '''
        Append an SLHA.Block, NamedBlock, Matrix or NamedMatrix to
        self.blocks or self.matrices depending on the type.
        '''
        block.preamble = preamble
        if isinstance(block,Matrix):
            self.matrices[block.name] = block
        else:
            self.blocks[block.name] = block

    def add_decay(self, decay, preamble=''):
        '''Append an SLHA.Decay to self.decays.'''
        decay.preamble = preamble
        self.decays[decay.PID] = decay

    def add_entry(self, blockname, key, value, name=None):
        '''
        Add a new field in a given block. If the Card instance already
        has such a block, a new field is appended. If the Card doesn't, a new
        block is created.
        '''
        if self.has_block(blockname):
            self.blocks[blockname].new_entry(key, value, name=name)
        elif self.has_matrix(blockname):
            self.matrices[blockname].new_entry(key, value, name=name)
        else:
            if type(key) is tuple:
                container = Matrix if name is None else NamedMatrix
                cplxcontainer = CMatrix if name is None else NamedMatrix
            else:
                container = Block if name is None else NamedBlock
                cplxcontainer = CBlock if name is None else NamedBlock


            if type(value) is complex:
                reblock = container(name=blockname)
                imblock = container(name='IM'+blockname)
                theblock = cplxcontainer(reblock, imblock)
            else:
                theblock = container(name=blockname)

            if name is None:
                theblock.new_entry(key, value)
            else:
                theblock.new_entry(key, value, name=name)

            self.add_block(theblock)

    def add_channel(self, PID, channel, partial, comment=None):
        '''
        Add a new channel in a given decay. The Decay for the given PID
        must already exist in the Card instance.
        '''
        assert(self.has_decay(PID)),('Tried adding a decay channel '
                                     'for non existent decay PID')
        self.decays[PID].new_channel(channel, partial, comment=comment)

    def new_channel(self, PIDs, BR, comment=None):
        '''Append a new channel to a given Decay block.'''
        self[PIDs]=BR
        if comment is not None: self._comments[PIDs]=comment

    def has_block(self, name):
        return name.lower() in self.blocks

    def has_matrix(self, name):
        return name.lower() in self.matrices

    def has_decay(self, PID):
        return PID in self.decays

    def write(self, filename, blockorder = [], preamble='', postamble=''):
        '''
        Write contents of Card in SLHA formatted style to "filename".
        Makes use of the __str__ methods written for the SLHA elements.
        Keyword arguments:
            blockorder - Specify an ordering for printing the block names.
                         Names given in blockorder will be printed first in the
                         order specified and others will be printed as ordered
                         in self.blocks
            preamble - Some text to write before the Block and Decay information
            postamble - Some text to write after the Block and Decay information
        '''

        with open(filename,'w') as out:
            out.write(preamble)

            blockorder = [x.lower() for x in blockorder]

            allblocks = list(self.blocks.keys()) + list(self.matrices.keys())

            other_blocks = [b for b in allblocks if b.lower() not in blockorder]

            for block in blockorder:
                if self.has_block(block):
                    out.write(str(self.blocks[block]))
                elif self.has_matrix(block):
                    out.write(str(self.matrices[block]))

            for block in other_blocks:
                if self.has_block(block):
                    out.write(str(self.blocks[block]))
                elif self.has_matrix(block):
                    out.write(str(self.matrices[block]))

            for decay in list(self.decays.values()):
                out.write(str(decay))
            out.write(postamble)

    def set_complex(self):
        '''
        Processes self.blocks and self.matrices to find real and imaginary
        pairs, upgrading them to a single complex version.
        '''

        cplx = []
        for imkey in list(self.blocks.keys()):
            if imkey.lower().startswith('im'):
                rekey = imkey[2:]
                if rekey in self.blocks:
                    if not isinstance(self.blocks[rekey], CBlock):
                        cplx.append((rekey,imkey))

        for rekey, imkey in cplx:
            reblk = self.blocks.get(rekey, None)
            imblk = self.blocks.pop(imkey, None)
            container = type(reblk)
            if container is Block:
                ctype = CBlock
            elif container is NamedBlock:
                ctype = CNamedBlock

            cblk = ctype(reblk, imblk)
            self.blocks[rekey] = cblk

        cplx = []
        for imkey in list(self.matrices.keys()):
            if imkey.lower().startswith('im'):
                rekey = imkey[2:]
                if rekey in self.matrices:
                    if not isinstance(self.matrices[rekey], CBlock):
                        cplx.append((rekey,imkey))

        for rekey, imkey in cplx:
            reblk = self.matrices.get(rekey, None)
            imblk = self.matrices.pop(imkey, None)

            container = type(reblk)
            if container is Matrix:
                ctype = CMatrix
            elif container is NamedMatrix:
                ctype = CNamedMatrix

            cblk = ctype(reblk, imblk)
            self.matrices[rekey] = cblk

class SLHAError(Exception): # Custom error name
    pass

def read(card, set_cplx=True):
    '''
    SLHA formatted card reader. Blocks and Decay structures are read
    into an SLHA.Card object which is returned. Comments are specified by the #
    character. Comments to the right of the structures as wel as the individual
    fields are stored.
    '''

    def get_comment(line):
        '''
        Returns characters following a "#" in line.
        returns empty string if match is not found.
        '''
        match = re.match(r'.*#\s*(.*)\s*', line)
        if match: return match.group(1).strip()
        else: return ''

    thecard = Card()
    pcard = open(card,'r')
    lines = iter(pcard)
    counter = 0

    try:
        while True:
            counter+=1 # keep track of line number
            stop=False
            try: ll=(last_line.strip())
            except NameError: ll = (next(lines)).strip()

            if not ll: continue

            first_chars = re.match(r'\s*(\S+).*',ll).group(1).lower()

            if first_chars=='block':
                try:
                    block_details = re.match(r'\s*block\s+([^\n]+)',ll,
                                             re.IGNORECASE).group(1)
                except AttributeError:
                    err = ('Invalid block format encountered ' +
                          'in line {} of {}.'.format(counter, card) )
                    raise SLHAError(err)

                bname = block_details.split('#')[0].strip()
                match_scale = re.match(r'(.*)\s+Q\s*=\s*([\deE\.\+\-]+)',bname,
                                         re.IGNORECASE)
                if match_scale is not None:
                    bname = match_scale.group(1)
                    scale = float(match_scale.group(2))
                else:
                    scale = None

                # comment = get_comment(ll)

                block_data, last_line, stop = read_until(lines,'block','decay')

                elements = []

                for datum in block_data:
                    counter +=1
                    # is_comment = re.match(r'\s*#.*',datum)

                    if not datum.strip(): continue
                    # if (not datum.strip() or is_comment): continue

                    info = re.match(r'\s*((?:\d+\s+)+)\s*(\S+).*',
                                    datum)
                    if not info:
                        log.info(datum)
                        log.info(('Ignored datum in block '+
                                '{},'.format(theblock.name) +
                                ' (line {} of {})'.format(counter, card)))
                        continue

                    key, value = info.group(1), info.group(2)

                    try:
                        key = int(key)
                    except ValueError:
                        key = tuple( map(int, key.split()) )

                    # dname = get_comment(datum)

                    try:
                        entry = float(value)
                    except ValueError:
                        entry = value
                    finally:
                        elements.append((key,value))#,dname))

                if not elements:
                    continue
                if type(elements[0][0]) is tuple:
                    theblock = Matrix(name=bname, scale=scale)
                else:
                    theblock = Block(name=bname, scale=scale)

                for k, v in elements:
                    theblock[k] = v

                thecard.add_block(theblock)

            elif first_chars=='decay':
                try:
                    decay_details = re.match(r'\s*decay\s+(.+)',ll,
                                             re.IGNORECASE).group(1)
                except AttributeError:
                    err = ('Invalid decay format encountered' +
                          'in line {} of {}.'.format(counter, card) )
                    raise SHLAReadError(err)

                info = re.match(r'\s*(\d+)\s+(\S+)\s+.*', decay_details)
                PID, total = info.group(1), info.group(2)

                comment = get_comment(decay_details)

                thedecay = Decay(PID=int(PID), total=float(total),
                                 comment=comment)

                decay_data, last_line, stop = read_until(lines,'block','decay')

                for datum in decay_data:
                    counter +=1
                    is_comment = re.match(r'\s*#.*',datum)

                    if ((not datum.strip()) or (is_comment)): continue

                    info = re.match(r'\s*(\S+)\s+(\d+)\s+(.+)',datum)
                    if not info:
                        log.info(datum)
                        log.info(('Ignored above datum in decay '+
                                '{},'.format(thedecay.PID) +
                                ' (line {} of {})'.format(counter, card)))
                        continue

                    BR, nout = info.group(1), info.group(2)
                    PIDinfo = info.group(3).split('#')[0]

                    PIDs = tuple( map(int, PIDinfo.split()) )

                    if len(PIDs)!=int(nout):
                        print ("Number of external particles in column 2 doesn't "
                               "match number of subsequent columns:")
                        log.info(datum)
                        log.info(('Ignored above datum in decay '+
                                '{},'.format(thedecay.PID) +
                                ' (line {} of {})'.format(counter, card)))
                        continue

                    comment = get_comment(datum)

                    thedecay.new_channel(PIDs, float(BR), comment=comment)

                thecard.add_decay(thedecay)

            if stop: raise StopIteration

    except StopIteration:
        pcard.close()

    if set_cplx:
        thecard.set_complex()

    return thecard


def read_until(lines, here, *args):
    '''
    Loops through an iterator of strings by calling next() until
    it reaches a line starting with a particular string.
    Case insensitive.
    Args:
        lines - iterator of strings
        here - string (plus any further argumnts).
               Reading will end if the line matches any of these.
    Return:
        lines_read - list of lines read
        line - last line that was read (containing string "here")
    '''
    end_strings = [here.lower()]+[a.lower() for a in args]
    lines_read = []
    line = ''
    stopiter = False
    while not any([line.strip().lower().startswith(x) for x in end_strings]):
        try:
            line = next(lines)
            lines_read.append(line.strip('\n'))
        except StopIteration:
            stopiter=True
            break
    try:
        if stopiter:
            return lines_read, '', stopiter
        else:
            return lines_read[:-1], lines_read[-1], stopiter
    except IndexError:
        return [],'',stopiter

if __name__=='__main__':
    pass



import unittest
import numpy as np
from . import amplitude, rge, observables
from math import sin, asin
from flavio.physics.eft import WilsonCoefficients
from flavio import Observable
from flavio.parameters import default_parameters
import copy
import flavio

s = 1.519267515435317e+24

c = copy.deepcopy(default_parameters)
par = c.get_central_all()

wc_obj = WilsonCoefficients()
wc_B0 = wc_obj.get_wc('bdbd', 4.2, par)
wc_Bs = wc_obj.get_wc('bsbs', 4.2, par)
wc_K = wc_obj.get_wc('sdsd', 2, par)

# this is the DeltaF=2 evolution matrix from mt to 4.2 GeV as obtained
# from the formulae in hep-ph/0102316
U_mb = np.array([[ 0.83693251,  0.        ,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
       [ 0.        ,  0.83693251,  0.        ,  0.        ,  0.        ,
         0.        ,  0.        ,  0.        ,  0.        ,  0.        ],
       [ 0.        ,  0.        ,  0.91882327,  0.        ,  0.        ,
         0.        , -0.04335413,  0.        ,  0.        ,  0.        ],
       [ 0.        ,  0.        ,  0.        ,  0.91882327,  0.        ,
         0.        ,  0.        , -0.04335413,  0.        ,  0.        ],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  1.68277548,
         0.        ,  0.        ,  0.        ,  2.06605957,  0.        ],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
         1.68277548,  0.        ,  0.        ,  0.        ,  2.06605957],
       [ 0.        ,  0.        , -0.92932569,  0.        ,  0.        ,
         0.        ,  2.31878868,  0.        ,  0.        ,  0.        ],
       [ 0.        ,  0.        ,  0.        , -0.92932569,  0.        ,
         0.        ,  0.        ,  2.31878868,  0.        ,  0.        ],
       [ 0.        ,  0.        ,  0.        ,  0.        , -0.00686759,
         0.        ,  0.        ,  0.        ,  0.53745488,  0.        ],
       [ 0.        ,  0.        ,  0.        ,  0.        ,  0.        ,
        -0.00686759,  0.        ,  0.        ,  0.        ,  0.53745488]]).reshape((10,10))[[0,4,8,1,5,9,2,6],:][:,[0,4,8,1,5,9,2,6]]

class TestMesonMixing(unittest.TestCase):
    def test_bmixing(self):
        # just some trivial tests to see if calling the functions raises an error
        m12d = amplitude.M12_d(par, wc_B0, 'B0')
        m12s = amplitude.M12_d(par, wc_Bs, 'Bs')
        # check whether order of magnitudes of SM predictions are right
        ps = 1e-12*s
        self.assertAlmostEqual(observables.DeltaM(wc_obj, par, 'B0')*ps, 0.55, places=0)
        self.assertAlmostEqual(observables.DeltaM(wc_obj, par, 'Bs')*ps, 18, places=-1)
        self.assertAlmostEqual(observables.DeltaGamma(wc_obj, par, 'B0')/0.00261*ps, 1, places=0)
        self.assertAlmostEqual(observables.DeltaGamma(wc_obj, par, 'Bs')/0.088*ps, 1, places=0)
        self.assertAlmostEqual(observables.a_fs(wc_obj, par, 'B0')/-4.7e-4, 1, places=0)
        self.assertAlmostEqual(observables.a_fs(wc_obj, par, 'Bs')/2.22e-5, 1, places=0)
        self.assertAlmostEqual(observables.S_BJpsiK(wc_obj, par), 0.73, places=1)
        self.assertAlmostEqual(observables.S_Bspsiphi(wc_obj, par), asin(+0.038), places=2)

    def test_bmixing_classes(self):
        ps = 1e-12*s
        self.assertAlmostEqual(Observable.get_instance('DeltaM_d').prediction_central(c, wc_obj)*ps, 0.53, places=0)
        self.assertAlmostEqual(Observable.get_instance('DeltaM_s').prediction_central(c, wc_obj)*ps, 18, places=-1)
        self.assertAlmostEqual(Observable.get_instance('DeltaGamma_d').prediction_central(c, wc_obj)/0.00261*ps, 1, places=-1)
        self.assertAlmostEqual(Observable.get_instance('DeltaGamma_s').prediction_central(c, wc_obj)/0.088*ps, 1, places=-1)
        self.assertAlmostEqual(Observable.get_instance('a_fs_d').prediction_central(c, wc_obj)/-4.7e-4, 1, places=-1)
        self.assertAlmostEqual(Observable.get_instance('a_fs_s').prediction_central(c, wc_obj)/2.22e-5, 1, places=-1)
        self.assertAlmostEqual(Observable.get_instance('S_psiK').prediction_central(c, wc_obj), 0.73, places=-1)
        self.assertAlmostEqual(Observable.get_instance('S_psiphi').prediction_central(c, wc_obj), asin(+0.038), places=-1)

    def test_running(self):
        c_in = np.array([ 0.20910694,  0.77740198,  0.54696337,  0.46407456,  0.42482153,
        0.95717777,  0.62733321,  0.87053086])
        c_out = flavio.physics.running.running.get_wilson(par, c_in, wc_obj.rge_derivative['bsbs'], 173.3, 4.2)
        c_out_U = np.dot(U_mb, c_in)
        np.testing.assert_almost_equal(c_out/c_out_U, np.ones(8), decimal=2)
        # compare eta at 2 GeV to the values in table 2 of hep-ph/0102316
        par_bju = par.copy()
        par_bju['alpha_s'] = 0.118
        par_bju['m_b'] = 4.4
        c_out_bju = rge.run_wc_df2(par_bju, c_in, 166., 2)
        self.assertAlmostEqual(c_out_bju[0]/c_in[0], 0.788, places=2)



r"""Functions for exclusive $B\to P\ell^+\ell^-$ decays."""

import flavio
from math import sqrt,pi
from flavio.physics.bdecays.common import lambda_K, beta_l, meson_quark, meson_ff
from flavio.physics import ckm
from flavio.classes import AuxiliaryQuantity
from flavio.config import config
from flavio.physics.running import running
from flavio.physics.common import conjugate_par, conjugate_wc, add_dict
from flavio.physics.bdecays import matrixelements, angular
from flavio.physics.bdecays.wilsoncoefficients import get_wceff, get_wceff_lfv, wctot_dict
from flavio.classes import Observable, Prediction
import warnings

def prefactor(q2, par, B, P):
    GF = par['GF']
    scale = config['renormalization scale']['bpll']
    alphaem = running.get_alpha(par, scale)['alpha_e']
    di_dj = meson_quark[(B,P)]
    xi_t = ckm.xi('t',di_dj)(par)
    return 4*GF/sqrt(2)*xi_t*alphaem/(4*pi)

# form factors
def get_ff(q2, par, B, P):
    ff_name = meson_ff[(B,P)] + ' form factor'
    return AuxiliaryQuantity.get_instance(ff_name).prediction(par_dict=par, wc_obj=None, q2=q2)

# get subleading hadronic contribution
def get_subleading(q2, wc_obj, par_dict, B, P, lep, cp_conjugate):
    if q2 <= 9:
        sub_name = B+'->'+P + 'll subleading effects at low q2'
        return AuxiliaryQuantity.get_instance(sub_name).prediction(par_dict=par_dict, wc_obj=wc_obj, q2=q2, cp_conjugate=cp_conjugate)
    elif q2 > 14:
        sub_name = B+'->'+P + 'll subleading effects at high q2'
        return AuxiliaryQuantity.get_instance(sub_name).prediction(par_dict=par_dict, wc_obj=wc_obj, q2=q2, cp_conjugate=cp_conjugate)
    else:
        return {}


def helicity_amps_ff(q2, wc_obj, par_dict, B, P, l1, l2, cp_conjugate):
    par = par_dict.copy()
    if cp_conjugate:
        par = conjugate_par(par)
    scale = config['renormalization scale']['bpll']
    label = meson_quark[(B,P)] + l1 + l2 # e.g. bsmumu, bdtautau
    if l1 == l2:
        # include SM contributions for LF conserving decay
        wc = wctot_dict(wc_obj, label, scale, par)
    else:
        wc = wc_obj.get_wc(label, scale, par)
    if cp_conjugate:
        wc = conjugate_wc(wc)
    if l1 == l2:
        wc_eff = get_wceff(q2, wc, par, B, P, l1, scale)
    else:
        wc_eff = get_wceff_lfv(q2, wc, par, B, P, l1, l2, scale)
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    mB = par['m_'+B]
    mP = par['m_'+P]
    mb = running.get_mb(par, scale)
    N = prefactor(q2, par, B, P)
    ff = get_ff(q2, par, B, P)
    h = angular.helicity_amps_p(q2, mB, mP, mb, 0, ml1, ml2, ff, wc_eff, N)
    return h

def helicity_amps(q2, wc_obj, par, B, P, l1, l2):
    if l1 != l2:
        return helicity_amps_ff(q2, wc_obj, par, B, P, l1, l2, cp_conjugate=False)
    else:
        if q2 >= 8.7 and q2 < 14:
            warnings.warn("The predictions in the region of narrow charmonium resonances are not meaningful")
        return add_dict((
            helicity_amps_ff(q2, wc_obj, par, B, P, l1, l2, cp_conjugate=False),
            get_subleading(q2, wc_obj, par, B, P, l1, cp_conjugate=False)
            ))

def helicity_amps_bar(q2, wc_obj, par, B, P, l1, l2):
    if l1 != l2:
        return helicity_amps_ff(q2, wc_obj, par, B, P, l1, l2, cp_conjugate=True)
    else:
        if q2 >= 8.7 and q2 < 14:
            warnings.warn("The predictions in the region of narrow charmonium resonances are not meaningful")
        return add_dict((
            helicity_amps_ff(q2, wc_obj, par, B, P, l1, l2, cp_conjugate=True),
            get_subleading(q2, wc_obj, par, B, P, l1, cp_conjugate=True)
            ))

def bpll_obs(function, q2, wc_obj, par, B, P, l1, l2):
    ml1 = par['m_'+l1]
    ml2 = par['m_'+l2]
    mB = par['m_'+B]
    mP = par['m_'+P]
    if q2 <= (ml1+ml2)**2 or q2 > (mB-mP)**2:
        return 0
    scale = config['renormalization scale']['bpll']
    mb = running.get_mb(par, scale)
    h     = helicity_amps(q2, wc_obj, par, B, P, l1, l2)
    h_bar = helicity_amps_bar(q2, wc_obj, par, B, P, l1, l2)
    J     = angular.angularcoeffs_general_p(h, q2, mB, mP, mb, 0, ml1, ml2)
    J_bar = angular.angularcoeffs_general_p(h_bar, q2, mB, mP, mb, 0, ml1, ml2)
    return function(J, J_bar)

def dGdq2(J):
    return 2 * (J['a'] + J['c']/3.)

def AFB_num(J):
    return J['b']

def FH_num(J):
    return 2 * (J['a'] + J['c'])

def dGdq2_cpaverage(J, J_bar):
    return (dGdq2(J) + dGdq2(J_bar))/2.

def AFB_cpaverage_num(J, J_bar):
    return (AFB_num(J) + AFB_num(J_bar))/2.

def FH_cpaverage_num(J, J_bar):
    return (FH_num(J) + FH_num(J_bar))/2.


# denominator of normalized observables
def denominator(J, J_bar):
    return 2*dGdq2_cpaverage(J, J_bar)

def bpll_obs_int(function, q2min, q2max, wc_obj, par, B, P, l1, l2):
    def obs(q2):
        return bpll_obs(function, q2, wc_obj, par, B, P, l1, l2)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)


def bpll_dbrdq2(q2, wc_obj, par, B, P, l1, l2):
    tauB = par['tau_'+B]
    return tauB * bpll_obs(dGdq2_cpaverage, q2, wc_obj, par, B, P, l1, l2)

def bpll_dbrdq2_int(q2min, q2max, wc_obj, par, B, P, l1, l2):
    def obs(q2):
        return bpll_dbrdq2(q2, wc_obj, par, B, P, l1, l2)
    return flavio.math.integrate.nintegrate(obs, q2min, q2max)/(q2max-q2min)

# Functions returning functions needed for Prediction instances

def bpll_dbrdq2_int_func(B, P, l1, l2):
    def fct(wc_obj, par, q2min, q2max):
        return bpll_dbrdq2_int(q2min, q2max, wc_obj, par, B, P, l1, l2)
    return fct

def bpll_dbrdq2_tot_func(B, P, l1, l2):
    def fct(wc_obj, par):
        mB = par['m_'+B]
        mP = par['m_'+P]
        ml1 = par['m_'+l1]
        ml2 = par['m_'+l2]
        q2max = (mB-mP)**2
        q2min = (ml1+ml2)**2
        return bpll_dbrdq2_int(q2min, q2max, wc_obj, par, B, P, l1, l2)
    return fct


def bpll_dbrdq2_func(B, P, l1, l2):
    def fct(wc_obj, par, q2):
        return bpll_dbrdq2(q2, wc_obj, par, B, P, l1, l2)
    return fct

def bpll_obs_int_ratio_func(func_num, func_den, B, P, l1, l2):
    def fct(wc_obj, par, q2min, q2max):
        num = bpll_obs_int(func_num, q2min, q2max, wc_obj, par, B, P, l1, l2)
        if num == 0:
            return 0
        den = bpll_obs_int(func_den, q2min, q2max, wc_obj, par, B, P, l1, l2)
        return num/den
    return fct

def bpll_obs_int_ratio_leptonflavour(func, B, P, lnum, lden):
    def fct(wc_obj, par, q2min, q2max):
        num = bpll_obs_int(func, q2min, q2max, wc_obj, par, B, P, lnum, lnum)
        if num == 0:
            return 0
        denom = bpll_obs_int(func, q2min, q2max, wc_obj, par, B, P, lden, lden)
        return num/denom
    return fct


def bpll_obs_ratio_func(func_num, func_den, B, P, l1, l2):
    def fct(wc_obj, par, q2):
        num = bpll_obs(func_num, q2, wc_obj, par, B, P, l1, l2)
        if num == 0:
            return 0
        den = bpll_obs(func_den, q2, wc_obj, par, B, P, l1, l2)
        return num/den
    return fct

# Observable and Prediction instances

_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
_observables = {
'AFB': {'func_num': AFB_cpaverage_num, 'tex': r'A_\text{FB}', 'desc': 'forward-backward asymmetry'},
'FH': {'func_num': FH_cpaverage_num, 'tex': r'F_H', 'desc': 'flat term'},
}
_hadr = {
'B0->K': {'tex': r"B^0\to K^0", 'B': 'B0', 'P': 'K0', },
'B+->K': {'tex': r"B^\pm\to K^\pm ", 'B': 'B+', 'P': 'K+', },
}
_hadr_lfv = {
'B0->K': {'tex': r"\bar B^0\to \bar K^0", 'B': 'B0', 'P': 'K0', },
'B+->K': {'tex': r"B^-\to K^-", 'B': 'B+', 'P': 'K+', },
'B0->pi': {'tex': r"\bar B^0\to \pi^0", 'B': 'B0', 'P': 'pi0', },
'B+->pi': {'tex': r"B^-\to \pi^-", 'B': 'B+', 'P': 'pi+', },
}
_tex_lfv = {'emu': r'e^+\mu^-', 'mue': r'\mu^+e^-',
    'taue': r'\tau^+e^-', 'etau': r'e^+\tau^-',
    'taumu': r'\tau^+\mu^-', 'mutau': r'\mu^+\tau^-'}

for l in ['e', 'mu', 'tau']:
    for M in _hadr.keys():
        for obs in sorted(_observables.keys()):
            _obs_name = "<" + obs + ">("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description('Binned ' + _observables[obs]['desc'] + r" in $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
            _obs.tex = r"$\langle " + _observables[obs]['tex'] + r"\rangle(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
            Prediction(_obs_name, bpll_obs_int_ratio_func(_observables[obs]['func_num'], denominator, _hadr[M]['B'], _hadr[M]['P'], l, l))

            _obs_name = obs + "("+M+l+l+")"
            _obs = Observable(name=_obs_name, arguments=['q2'])
            _obs.set_description(_observables[obs]['desc'][0].capitalize() + _observables[obs]['desc'][1:] + r" in $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
            _obs.tex = r"$" + _observables[obs]['tex'] + r"(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
            Prediction(_obs_name, bpll_obs_ratio_func(_observables[obs]['func_num'], denominator, _hadr[M]['B'], _hadr[M]['P'], l, l))

        # binned branching ratio
        _obs_name = "<dBR/dq2>("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
        _obs.set_description(r"Binned differential branching ratio of $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
        _obs.tex = r"$\langle \frac{d\text{BR}}{dq^2} \rangle(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
        Prediction(_obs_name, bpll_dbrdq2_int_func(_hadr[M]['B'], _hadr[M]['P'], l, l))

        # differential branching ratio
        _obs_name = "dBR/dq2("+M+l+l+")"
        _obs = Observable(name=_obs_name, arguments=['q2'])
        _obs.set_description(r"Differential branching ratio of $" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-$")
        _obs.tex = r"$\frac{d\text{BR}}{dq^2}(" + _hadr[M]['tex'] +_tex[l]+r"^+"+_tex[l]+"^-)$"
        Prediction(_obs_name, bpll_dbrdq2_func(_hadr[M]['B'], _hadr[M]['P'], l, l))

# Lepton flavour ratios
for l in [('mu','e'), ('tau','mu'),]:
    for M in _hadr.keys():

            # binned ratio of BRs
            _obs_name = "<R"+l[0]+l[1]+">("+M+"ll)"
            _obs = Observable(name=_obs_name, arguments=['q2min', 'q2max'])
            _obs.set_description(r"Ratio of partial branching ratios of $" + _hadr[M]['tex'] +_tex[l[0]]+r"^+ "+_tex[l[0]]+r"^-$" + " and " + r"$" + _hadr[M]['tex'] +_tex[l[1]]+r"^+ "+_tex[l[1]]+"^-$")
            _obs.tex = r"$\langle R_{" + _tex[l[0]] + ' ' + _tex[l[1]] + r"} \rangle(" + _hadr[M]['tex'] + r"\ell^+\ell^-)$"
            Prediction(_obs_name, bpll_obs_int_ratio_leptonflavour(dGdq2_cpaverage, _hadr[M]['B'], _hadr[M]['P'], *l))

# Lepton flavour violating decays
for ll in [('e','mu'), ('mu','e'), ('e','tau'), ('tau','e'), ('mu','tau'), ('tau','mu')]:
    for M in _hadr_lfv:
        for br in ['BR',]:
            _obs_name = br + "("+M+''.join(ll)+")"
            _obs = Observable(_obs_name)
            _obs.set_description(r"Total branching ratio of $"+_hadr_lfv[M]['tex']+' '+_tex_lfv[''.join(ll)]+r"$")
            _obs.tex = r"$\text{BR}(" + _hadr_lfv[M]['tex']+' '+_tex_lfv[''.join(ll)]+r")$"
            Prediction(_obs_name, bpll_dbrdq2_tot_func(_hadr_lfv[M]['B'], _hadr_lfv[M]['P'], ll[0], ll[1]))



import numpy as np
import flavio

try:
    import emcee
except:
    pass

class emceeScan(object):
    """
    """

    def __init__(self, fit, nwalkers=None, **kwargs):

        assert isinstance(fit, flavio.statistics.fits.BayesianFit), "emcee fit object must be an instance of BayesianFit"
        self.fit = fit

        self.dimension = len(fit.get_random)
        if nwalkers is None:
            self.nwalkers = self.dimension * 10
        else:
            self.nwalkers = nwalkers
        self.start = [fit.get_random for i in range(self.nwalkers)]


        self.mc = emcee.EnsembleSampler(nwalkers=self.nwalkers,
                                        dim=self.dimension,
                                        lnpostfn=self.fit.log_target,
                                        **kwargs)

    def run(self, steps, burnin=1000):
        pos = self.start
        if burnin > 0:
            pos, prob, state = self.mc.run_mcmc(self.start, burnin)
        self.mc.reset()
        self.mc.run_mcmc(pos, steps)

    @property
    def result(self):
        return self.mc.flatchain[:]

    def save_result(self, file):
        res = self.result
        np.save(file, res)



import unittest
import numpy as np
import flavio
from flavio.physics.bdecays.formfactors.b_v import bsz_parameters
from flavio.physics.bdecays.bvll import observables, observables_bs
import cmath

class TestBVll(unittest.TestCase):
    def test_compare_to_Davids_old_code(self):
        # this actually works to 3% and better! Still, setting it to 15%
        # here to not give any error in case of future changes to the central value of the code
        delta = 0.15
        self.assertAlmostEqual(flavio.sm_prediction("AFB(B0->K*mumu)", q2=1)/-0.15349451759556973, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("FL(B0->K*mumu)", q2=1)/0.6406629242990924, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S5(B0->K*mumu)", q2=1)/0.22071783227496475, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P4p(B0->K*mumu)", q2=1)/0.15140284164390483, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P5p(B0->K*mumu)", q2=1)/0.5064923738851054, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("AFB(B0->K*mumu)", q2=6)/0.17920578512694066, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("FL(B0->K*mumu)", q2=6)/0.6746725598733865, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S4(B0->K*mumu)", q2=6)/-0.24476949751293892, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S5(B0->K*mumu)", q2=6)/-0.3735858481345886, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P4p(B0->K*mumu)", q2=6)/-0.529132300413707, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P5p(B0->K*mumu)", q2=6)/-0.807602014278806, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("AFB(B0->K*mumu)", q2=15)/0.406194755197377, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("FL(B0->K*mumu)", q2=15)/0.36807161547568185, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S3(B0->K*mumu)", q2=15)/-0.14872271241628923, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S4(B0->K*mumu)", q2=15)/-0.2919191029733921, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("S5(B0->K*mumu)", q2=15)/-0.34293705381921147, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P4p(B0->K*mumu)", q2=15)/-0.6065016593226283, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("P5p(B0->K*mumu)", q2=15)/-0.7124983944730857, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=1)/5.7131974656129313e-8, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=6)/5.524478678779278e-8, 1, delta=delta)
        self.assertAlmostEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=15)/7.412258847550239e-8, 1, delta=delta)
        # this is just a very rough comparison to the literature
        delta = 0.2
        self.assertAlmostEqual(flavio.sm_prediction("P1(B0->K*mumu)", q2=3), 0, delta=delta)
        delta = 0.1
        self.assertAlmostEqual(flavio.sm_prediction("P1(B0->K*mumu)", q2=17), -0.64, delta=delta)
        delta = 0.1
        self.assertAlmostEqual(flavio.sm_prediction("P2(B0->K*mumu)", q2=6)/(2/3.*0.179/0.325), 1, delta=delta)
        delta = 0.1
        self.assertAlmostEqual(flavio.sm_prediction("P2(B0->K*mumu)", q2=17)/0.37, 1, delta=delta)
        delta = 0.02
        self.assertAlmostEqual(flavio.sm_prediction("P3(B0->K*mumu)", q2=3), 0, delta=delta)
        delta = 0.02
        self.assertAlmostEqual(flavio.sm_prediction("P3(B0->K*mumu)", q2=17), 0, delta=delta)

    def test_unphysical(self):
        # check BR calculation yields zero outside kinematical limits
        self.assertEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=0.01), 0)
        self.assertEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=25), 0)
        # and also *at* kinemetical limits
        par = flavio.default_parameters.get_central_all()
        q2min = 4*par['m_mu']**2
        q2max = (par['m_B0']-par['m_K*0'])**2
        self.assertAlmostEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=q2min), 0, delta=1e-10)
        self.assertAlmostEqual(flavio.sm_prediction("dBR/dq2(B0->K*mumu)", q2=q2max), 0, delta=1e-10)

        # same for angular observables (make sure no division by 0)
        self.assertEqual(flavio.sm_prediction("S5(B0->K*mumu)", q2=0.01), 0)
        self.assertEqual(flavio.sm_prediction("S5(B0->K*mumu)", q2=25), 0)

    def test_bs_timedep(self):
        q2 = 3
        wc_obj = flavio.WilsonCoefficients()
        par = flavio.default_parameters.get_central_all()
        B = 'Bs'
        V = 'phi'
        l = 'mu'
        # a set of parameters with y_s=0!
        par_y0 = par.copy()
        par_y0['DeltaGamma/Gamma_Bs']=0

        # compare without lifetime effect: must be equal!
        self.assertEqual(
            observables.bvll_obs(     observables.dGdq2_ave,       q2, wc_obj, par_y0, B, V, l),
            observables_bs.bsvll_obs( observables_bs.dGdq2_ave_Bs, q2, wc_obj, par_y0, B, V, l))
        self.assertEqual(
            observables.bvll_obs(     observables.FL,       q2, wc_obj, par_y0, B, V, l),
            observables_bs.bsvll_obs( observables_bs.FL_Bs, q2, wc_obj, par_y0, B, V, l))
        for i in [3, 4, 7]: # S3,4,7
            self.assertEqual(
                observables.bvll_obs(     lambda J, J_bar:         observables.S_experiment(J, J_bar, i),               q2, wc_obj, par_y0, B, V, l),
                observables_bs.bsvll_obs( lambda y, J, J_bar, J_h: observables_bs.S_experiment_Bs(y, J, J_bar, J_h, i), q2, wc_obj, par_y0, B, V, l))

        # check that the phase phi has the right convention
        q_over_p = flavio.physics.mesonmixing.observables.q_over_p(wc_obj, par, B)
        phi = cmath.phase(-q_over_p) # the phase of q/p
        self.assertAlmostEqual(phi, 0.04, delta=0.01)

        # compare WITH lifetime effect: angular observables must be similar
        delta = 0.01
        self.assertAlmostEqual(
            observables.bvll_obs(     observables.FL,       q2, wc_obj, par, B, V, l)/
            observables_bs.bsvll_obs( observables_bs.FL_Bs, q2, wc_obj, par, B, V, l),
            1, delta=delta)
        for i in [4, 7]: # S4,7
            self.assertAlmostEqual(
                observables.bvll_obs(     lambda J, J_bar:         observables.S_experiment(J, J_bar, i),               q2, wc_obj, par, B, V, l)/
                observables_bs.bsvll_obs( lambda y, J, J_bar, J_h: observables_bs.S_experiment_Bs(y, J, J_bar, J_h, i), q2, wc_obj, par, B, V, l),
                1, delta=delta)
        for i in [3]: # S3: look at differnece only
            self.assertAlmostEqual(
                observables.bvll_obs(     lambda J, J_bar:         observables.S_experiment(J, J_bar, i),               q2, wc_obj, par, B, V, l) -
                observables_bs.bsvll_obs( lambda y, J, J_bar, J_h: observables_bs.S_experiment_Bs(y, J, J_bar, J_h, i), q2, wc_obj, par, B, V, l),
                0, delta=0.01)
        # compare WITH lifetime effect: BR suppressed by ~6%!
        self.assertAlmostEqual(
            observables.bvll_obs(     observables.dGdq2_ave,       q2, wc_obj, par, B, V, l)/
            observables_bs.bsvll_obs( observables_bs.dGdq2_ave_Bs, q2, wc_obj, par, B, V, l),
            1.06, delta=0.02)

        # and now just check a few observables to see if any errors are raised
        flavio.sm_prediction("FL(Bs->phimumu)", q2=1)
        flavio.sm_prediction("S3(Bs->phimumu)", q2=1)
        flavio.sm_prediction("S4(Bs->phimumu)", q2=1)
        flavio.sm_prediction("S7(Bs->phimumu)", q2=1)



import unittest
import numpy as np
from .measurements import *
from .classes import *


class TestMeasurements(unittest.TestCase):
    def test_measurements(self):
        m = Measurement.get_instance('CMS+LHCb Bq->mumu 2014')
        self.assertEqual(m.experiment, 'CMS+LHCb')



r"""Functions for imroved Isgur-Wise relations to obtain tensor
form factor $f_T$ from $f_+$ and $f_0$s in the heavy quark limit."""

import flavio
from math import pi,log



def improved_isgur_wise(process, q2, ff, par, scale):
    pd = flavio.physics.bdecays.formfactors.b_p.bcl.process_dict[process]
    mB = par['m_'+pd['B']]
    mP = par['m_'+pd['P']]
    mb = flavio.physics.running.running.get_mb(par, scale)
    alpha_s = flavio.physics.running.running.get_alpha(par, scale)['alpha_s']
    # eq. (3.8) of arXiv:1006.5013
    kappa = 1-2*alpha_s/(3*pi)*log(scale/mb)
    # eq. (3.6) of arXiv:1006.5013
    aT = par[process + ' IW a_T'] # a_T is a relative power correction
    ff['fT'] = kappa*ff['f+']*mB*(mB+mP)/q2 * (1 + aT)
    return ff



import numpy as np
import flavio

try:
    import pypmc
except:
    pass

class pypmcScan(object):
    """
    """

    def __init__(self, fit, **kwargs):

        assert isinstance(fit, flavio.statistics.fits.BayesianFit), "PyPMC fit object must be an instance of BayesianFit"
        self.fit = fit
        # start value is a random vector
        self.start = fit.get_random
        self.dimension = len(self.start)
        # generate another random vector to guess an initial step size
        _initial_sigma = np.absolute(self.start - fit.get_random)
        self._initial_proposal = pypmc.density.gauss.LocalGauss(_initial_sigma * np.eye(self.dimension))

        self.mc = pypmc.sampler.markov_chain.AdaptiveMarkovChain(target=self.fit.log_target,
                                                                 proposal=self._initial_proposal,
                                                                 start=self.start,
                                                                 save_target_values=True,
                                                                 **kwargs)

    def run(self, steps, burnin=1000, adapt=500):
        done = 0
        self.mc.run( burnin )
        self.mc.clear()
        while done < steps:
            todo = min(steps-done, adapt)
            accepted = self.mc.run( todo )
            done += todo
            self.mc.adapt()

    @property
    def find_burnin(self):
        target = self.mc.target_values[:] # this contains all log likelihood values of the chain
        target_end = target[-len(target)//10:] # take the last 10% of the chain
        target_end_mn = np.mean(target_end) # mean of the last 10%
        target_end_std = np.std(target_end) # std. deviation of the last 10%
        # find the index of the first entry where the log likelihood is greater
        # than the mean of the last 10% - 3 standard deviations
        first_good = np.argmax(target > target_end_mn - 3*target_end_std)
        return first_good

    @property
    def result(self):
        return self.mc.samples[:][self.find_burnin:]

    def save_result(self, file):
        res = self.result
        np.save(file, res)



r"""Functions for $B\to\ell\nu$ decays."""

import flavio
from math import pi

def br_plnu_general(wc, par, Vij, P, qiqj, lep, delta=0):
    r"""Branching ratio of general $P^+\to\ell^+\nu_\ell$ decay.

    `Vij` is the appropriate CKM matrix element.
    `delta` (detaults to 0) is a correction factor to account for different
    experimental treatment of electromagnetic effects, for instance.
    """
    ml = par['m_'+lep]
    mP = par['m_'+P]
    GF = par['GF']
    tau = par['tau_'+P]
    f = par['f_'+P]
    # Wilson coefficient dependence
    qqlnu = qiqj + lep + 'nu'
    rWC = (wc['CV_'+qqlnu] - wc['CVp_'+qqlnu]) + mP**2/ml * (wc['CS_'+qqlnu] - wc['CSp_'+qqlnu])
    N = tau * GF**2 * f**2 / (8*pi) * mP * ml**2  * (1 - ml**2/mP**2)**2
    return N * abs(Vij)**2 * abs(rWC)**2 * (1 + delta)

def br_blnu(wc_obj, par, lep):
    r"""Branching ratio of $B^+\to\ell^+\nu_\ell$."""
    # CKM element
    Vub = flavio.physics.ckm.get_ckm(par)[0,2]
    # renormalization scale
    scale = flavio.config['renormalization scale']['bll']
    # Wilson coefficients
    wc = wc_obj.get_wc('bu' + lep + 'nu', scale, par)
    # add SM contribution to Wilson coefficient
    wc['CV_bu'+lep+'nu'] += flavio.physics.bdecays.wilsoncoefficients.get_CVSM(par, scale, nf=5)
    return br_plnu_general(wc, par, Vub, 'B+', 'bu', lep, delta=0)

# function returning function needed for prediction instance
def br_blnu_fct(lep):
    def f(wc_obj, par):
        return br_blnu(wc_obj, par, lep)
    return f

# Observable and Prediction instances
_tex = {'e': 'e', 'mu': '\mu', 'tau': r'\tau'}
for l in ['e', 'mu', 'tau']:
    _obs_name = "BR(B+->"+l+"nu)"
    _obs = flavio.classes.Observable(_obs_name)
    _obs.set_description(r"Branching ratio of $B^+\to "+_tex[l]+r"^+\nu_"+_tex[l]+r"$")
    _obs.tex = r"$\text{BR}(B^+\to "+_tex[l]+r"^+\nu_"+_tex[l]+r")$"
    flavio.classes.Prediction(_obs_name, br_blnu_fct(l))



import numpy as np
import scipy.stats
import math


########## ProbabilityDistribution Class ##########
class ProbabilityDistribution(object):
   """Common base class for all probability distributions"""

   def __init__(self, central_value):
      self.central_value = central_value

   def get_central(self):
      return self.central_value

    # here we define the __hash__ and __eq__ methods to be able to use
    # instances as dictionary keys.

   def __hash__(self):
      return id(self)

   def __eq__(self, other):
      return id(self) == id(other)



class UniformDistribution(ProbabilityDistribution):

   def __init__(self, central_value, half_range):
      super().__init__(central_value)
      self.half_range = half_range
      self.range = (self.central_value - self.half_range,
                    self.central_value + self.half_range)

   def get_random(self, size=None):
      return np.random.uniform(self.range[0], self.range[1], size)

   def logpdf(self, x):
       if x < self.range[0] or x >= self.range[1]:
           return 0
       else:
           return -math.log(2*self.half_range)

class DeltaDistribution(ProbabilityDistribution):
   def __init__(self, central_value):
      super().__init__(central_value)

   def get_random(self, size=None):
      if size is None:
          return self.central_value
      else:
          return self.central_value * np.ones(size)

   def logpdf(self, x):
       if x == self.central_value:
           return 0.
       else:
           return -np.inf

class NormalDistribution(ProbabilityDistribution):

   def __init__(self, central_value, standard_deviation):
      super().__init__(central_value)
      self.standard_deviation = standard_deviation

   def get_random(self, size=None):
      return np.random.normal(self.central_value, self.standard_deviation, size)

   def logpdf(self, x):
       return scipy.stats.norm.logpdf(x, self.central_value, self.standard_deviation)

class AsymmetricNormalDistribution(ProbabilityDistribution):

   def __init__(self, central_value, right_deviation, left_deviation):
      super().__init__(central_value)
      if right_deviation < 0 or left_deviation < 0:
          raise ValueError("Left and right standard deviations must be positive numbers")
      self.right_deviation = right_deviation
      self.left_deviation = left_deviation

   def get_random(self, size=None):
        r = np.random.uniform()
        a = abs(self.left_deviation/(self.right_deviation+self.left_deviation))
        if  r > a:
            x = abs(np.random.normal(0,self.right_deviation))
            return self.central_value + x
        else:
            x = abs(np.random.normal(0,self.left_deviation))
            return self.central_value - x

   def logpdf(self, x):
       # values of the PDF at the central value
       p_right = scipy.stats.norm.pdf(self.central_value, self.central_value, self.right_deviation)
       p_left = scipy.stats.norm.pdf(self.central_value, self.central_value, self.left_deviation)
       if x < self.central_value:
           # left-hand side: scale factor
           r = 2*p_right/(p_left+p_right)
           return math.log(r) + scipy.stats.norm.logpdf(x, self.central_value, self.left_deviation)
       else:
           # left-hand side: scale factor
           r = 2*p_left/(p_left+p_right)
           return math.log(r) + scipy.stats.norm.logpdf(x, self.central_value, self.right_deviation)

class MultivariateNormalDistribution(ProbabilityDistribution):

   def __init__(self, central_value, covariance):
      super().__init__(central_value)
      assert np.all(np.linalg.eigvals(covariance) > 0), "The covariance matrix is not positive definite!" + str(covariance)
      self.covariance = covariance

   def get_random(self, size=None):
      return np.random.multivariate_normal(self.central_value, self.covariance, size)

   def logpdf(self, x):
       # to avoid ill-conditioned covariance matrices, all data are rescaled
       # by the inverse variances
       err = np.sqrt(np.diag(self.covariance))
       scale_matrix = np.outer(err, err)
       pdf_scaled = scipy.stats.multivariate_normal.logpdf(x/err, self.central_value/err, self.covariance/scale_matrix)
       return pdf_scaled + math.log(np.linalg.det(self.covariance/scale_matrix)/np.linalg.det(self.covariance))/2.




# Auxiliary functions

def combine_distributions(probability_distributions):
    """Combine a set of univariate probability distributions.

    This function is meant for combining uncertainties on a single parameter/
    observable. As an argument, it takes a list of Gaussians that all have
    the same mean. It returns their convolution, but with location equal to the
    original mean.

    This should be generalized to other kinds of distributions in the future...
    """
    # if there's just one: return it immediately
    if len(probability_distributions) == 1:
        return probability_distributions[0]
    central_value = probability_distributions[0].central_value # central value of the first dist
    assert isinstance(central_value, float), "Combination only implemented for univariate distributions"
    gaussians = []
    for pd in probability_distributions:
        if pd.central_value != central_value:
            raise ValueError("All distributions to be combined must have the same central value")
        if isinstance(pd, DeltaDistribution):
            # delta distributions (= no error) can be skipped
            continue
        elif isinstance(pd, NormalDistribution):
            gaussians.append(pd)
        else:
            raise ValueError("Combination only implemented for normal distributions")
    err_squared = sum([pd.standard_deviation**2 for pd in gaussians])
    return NormalDistribution(central_value=central_value, standard_deviation=math.sqrt(err_squared))



from math import sqrt
import numpy as np
from flavio.physics.bdecays.formfactors.common import z
from flavio.physics.bdecays.formfactors.b_p.isgurwise import improved_isgur_wise

def pole(ff, mres, q2):
    mresdict = {'f0': 0,'f+': 1,'fT': 1}
    m = mres[mresdict[ff]]
    if m == 0 or m is None:
        return 1
    return 1/(1-q2/m**2)

resonance_masses = {}
# resonance masses used in 1509.06235
# this is m_Bs*(0+), m_Bs*(1-)
resonance_masses['B->K'] = [5.711, 5.4154]
# resonance masses used in 1505.03925
resonance_masses['B->D'] = [6.420, 6.330]
# resonance masses used in 1503.07839v2
# this is m_B*(1-)
resonance_masses['B->pi'] = [None, 5.319]
# resonance masses used in 1501.05373v3
# this is m_B*(0+), m_B*(1-)
resonance_masses['Bs->K'] = [5.63, 5.3252]

# the following dict maps transitions to mesons. Note that it doesn't really
# matter whether the charged or neutral B/K/pi are used here. We don't
# distinguish between charged and neutral form factors anyway.
process_dict = {}
process_dict['B->K'] =    {'B': 'B0', 'P': 'K0',}
process_dict['Bs->K'] =    {'B': 'Bs', 'P': 'K+',}
process_dict['B->D'] =    {'B': 'B+', 'P': 'D0',}
process_dict['B->pi'] =   {'B': 'B+', 'P': 'pi0',}


def param_fplusT(mB, mP, a_i, q2, t0=None):
    Z = z(mB, mP, q2, t0)
    n = len(a_i)
    k = np.arange(n)
    return ( a_i * (Z**k - (-1)**(k - n) * k/n * Z**n) ).sum()

def param_f0(mB, mP, a_i, q2, t0=None):
    Z = z(mB, mP, q2, t0)
    k = np.arange(len(a_i))
    return ( a_i * Z**k ).sum()

def ff(process, q2, par, n=3, t0=None):
    r"""Central value of $B\to P$ form factors in the standard convention
    and BCL parametrization (arXiv:0807.2722).

    The standard convention defines the form factors $f_+$, $f_0$, and $f_T$.
    """
    pd = process_dict[process]
    mres = resonance_masses[process]
    mB = par['m_'+pd['B']]
    mP = par['m_'+pd['P']]
    ff = {}
    a={}
    for i in ['f+', 'fT', 'f0']:
        a[i] = [ par[process + ' BCL' + ' a' + str(j) + '_' + i] for j in range(n) ]
    ff['f+'] = pole('f+', mres, q2) * param_fplusT(mB, mP, a['f+'], q2, t0)
    ff['fT'] = pole('fT', mres, q2) * param_fplusT(mB, mP, a['fT'], q2, t0)
    ff['f0'] = pole('f0', mres, q2) * param_f0(mB, mP, a['f0'], q2, t0)
    return ff

def ff_isgurwise(process, q2, par, scale, n=3, t0=None):
    r"""Central value of $B\to P$ form factors in the standard convention
    and BCL parametrization (arXiv:0807.2722) for $f_0$ and $f_+$, but using
    an improved Isgur-Wise relation in the heavy quark limit for $f_T$.
    """
    pd = process_dict[process]
    mres = resonance_masses[process]
    mB = par['m_'+pd['B']]
    mP = par['m_'+pd['P']]
    ff = {}
    a={}
    for i in ['f+', 'f0']:
        a[i] = [ par[process + ' BCL' + ' a' + str(j) + '_' + i] for j in range(n) ]
    ff['f+'] = pole('f+', mres, q2) * param_fplusT(mB, mP, a['f+'], q2, t0)
    ff['f0'] = pole('f0', mres, q2) * param_f0(mB, mP, a['f0'], q2, t0)
    ff = improved_isgur_wise(process, q2, ff, par, scale=scale)
    return ff
